from typing import List, Sequence

import cytoolz as tz
import numpy as np
from multiprocess.pool import Pool
from tqdm import tqdm

from robustnessgym.core.constants import SLICEBUILDERS
from robustnessgym.core.slice import SliceDataPanel as DataPanel
from robustnessgym.slicebuilders.subpopulation import Subpopulation


class SubpopulationCollection(Subpopulation):
    def __init__(self, subpopulations: Sequence[Subpopulation], *args, **kwargs):

        super(SubpopulationCollection, self).__init__(
            identifiers=list(
                tz.concat(
                    [subpopulation.identifiers for subpopulation in subpopulations]
                )
            ),
            *args,
            **kwargs,
        )

        # TODO(karan): some subpopulations aren't compatible with each other (e.g.
        #  single column vs. multi column):
        #  add some smarter logic here to handle this

        # Store the subpopulations
        self.subpopulations = subpopulations

    def __call__(
        self,
        dp: DataPanel,
        columns: List[str],
        num_proc: int = None,
        *args,
        **kwargs,
    ):

        if not num_proc or num_proc == 1:
            slices = []
            slice_membership = []
            # Apply each slicebuilder in sequence
            for i, slicebuilder in tqdm(enumerate(self.subpopulations)):
                # Apply the slicebuilder
                slices_i, slice_membership_i = slicebuilder(
                    dp=dp,
                    columns=columns,
                    *args,
                    **kwargs,
                )

                # Add in the slices and slice membership
                slices.extend(slices_i)
                slice_membership.append(slice_membership_i)

        else:
            # TODO(karan): cleanup, make mp.Pool support simpler across the library
            with Pool(num_proc) as pool:
                slices, slice_membership = zip(
                    *pool.map(
                        lambda sb: sb(
                            dp=dp,
                            columns=columns,
                            *args,
                            **kwargs,
                        ),
                        [slicebuilder for slicebuilder in self.subpopulations],
                    )
                )

                # Combine all the slices
                slices = list(tz.concat(slices))

            def _store_updates(batch, indices):

                # Each Subpopulation will generate slices
                for i, subpopulation in enumerate(self.subpopulations):
                    updates = subpopulation.construct_updates(
                        slice_membership=slice_membership[i][indices],
                        columns=columns,
                    )

                    batch = subpopulation.store(
                        batch=batch,
                        updates=updates,
                    )

                return batch

            if isinstance(dp, DataPanel):
                dp = dp.map(
                    _store_updates,
                    with_indices=True,
                    batched=True,
                )

                for subpopulation in self.subpopulations:
                    # Update the DataPanel's history
                    dp.update_tape(
                        path=[SLICEBUILDERS, subpopulation.category],
                        identifiers=subpopulation.identifiers,
                        columns=columns,
                    )

        # Combine all the slice membership matrices
        slice_membership = np.concatenate(slice_membership, axis=1)

        return slices, slice_membership

    def apply(
        self,
        batch: DataPanel,
        columns: List[str],
        slice_membership: np.ndarray = None,
        *args,
        **kwargs,
    ) -> np.ndarray:

        # Each Subpopulation will generate slices
        for subpopulation, end_idx in zip(
            self.subpopulations, np.cumsum([s.num_slices for s in self.subpopulations])
        ):
            # Fill out the slice_membership
            slice_membership[
                :, end_idx - subpopulation.num_slices : end_idx
            ] = subpopulation.apply(
                slice_membership=slice_membership[
                    :, end_idx - subpopulation.num_slices : end_idx
                ],
                batch=batch,
                columns=columns,
            )

        return slice_membership

    # TODO(karan): add combinations for collections



from __future__ import annotations

from typing import List

import cytoolz as tz
import numpy as np
import tqdm

from robustnessgym.core.constants import GENERIC
from robustnessgym.core.slice import SliceDataPanel as DataPanel
from robustnessgym.slicebuilders.slicebuilder import SliceBuilder


class SliceBuilderCollection(SliceBuilder):
    """Collection of Slice Builders."""

    def __init__(self, slicebuilders: List[SliceBuilder], *args, **kwargs):
        super(SliceBuilderCollection, self).__init__(
            category=GENERIC,
            identifiers=list(
                tz.concat([slicebuilder.identifiers for slicebuilder in slicebuilders])
            ),
            *args,
            **kwargs,
        )

        # TODO(karan): some slicebuilders aren't compatible with each other (e.g.
        #  single column vs. multi column):
        #  add some smarter logic here to handle this

        # Store the subpopulations
        self.slicebuilders = slicebuilders

    def __repr__(self):
        # TODO(karan): format this nicely
        return (
            f"{self.__class__.__name__}("
            f"{[str(slicebuilder) for slicebuilder in self.slicebuilders]})]"
        )

    def __call__(
        self,
        dp: DataPanel,
        columns: List[str],
        mask: List[int] = None,
        store_compressed: bool = None,
        store: bool = None,
        *args,
        **kwargs,
    ):

        if mask:
            raise NotImplementedError(
                "Mask not supported for SliceBuilderCollection yet."
            )

        slices = []
        slice_membership = []

        # Apply each slicebuilder in sequence
        for i, slicebuilder in tqdm.tqdm(enumerate(self.slicebuilders)):
            # Apply the slicebuilder
            dp, slices_i, slice_membership_i = slicebuilder(
                batch_or_dataset=dp,
                columns=columns,
                mask=mask,
                store_compressed=store_compressed,
                store=store,
                *args,
                **kwargs,
            )

            # Add in the slices and slice membership
            slices.extend(slices_i)
            slice_membership.append(slice_membership_i)

        slice_membership = np.concatenate(slice_membership, axis=1)

        return dp, slices, slice_membership



from collections import OrderedDict
from typing import List, Union

from datasets.features import ClassLabel, Sequence, Value

from robustnessgym.core.identifier import Identifier
from robustnessgym.core.slice import SliceDataPanel as DataPanel
from robustnessgym.tasks.schema import Schema


class Task:
    """Class for tasks in Robustness Gym."""

    dataset_to_task = {}

    def __init__(
        self,
        identifier: Union[str, Identifier] = "GenericTask",
        input_schema: Schema = None,
        output_schema: Schema = None,
        metrics: List[str] = None,
        *args,
        **kwargs,
    ):
        self._identifier = identifier
        self._input_schema = input_schema
        self._output_schema = output_schema
        self._metrics = metrics

    def __repr__(self):
        return (
            f"task: {self._identifier}\n\n"
            f"Input: {str(self._input_schema)}\n\n"
            f"Output: {str(self._output_schema)}"
        )

    @classmethod
    def lookup(cls, dataset: str):
        return cls.dataset_to_task[dataset]

    @classmethod
    def list_datasets(cls):
        return []

    @property
    def identifier(self):
        """Task identifier."""
        return self._identifier

    @property
    def metrics(self):
        """Task metrics."""
        return self._metrics

    @property
    def input_schema(self):
        """Input schema for the task."""
        return self._input_schema

    @property
    def output_schema(self):
        """Output schema for the task."""
        return self._output_schema

    # @classmethod
    # def from_identifier(cls, identifier):
    #     return getattr()

    @classmethod
    def create(cls, task: str):
        # TODO(karan): figure out how to getattr this
        if task == "TernaryNaturalLanguageInference":
            return TernaryNaturalLanguageInference()
        else:
            raise NotImplementedError

    def remap_schema(self, dp: DataPanel):
        # Ground the schema to the dp
        input_grounding, reversed_input_grounding = self.input_schema.ground(
            dp.features
        )
        output_grounding, reversed_output_grounding = self.output_schema.ground(
            dp.features
        )

        for col in self.input_schema.columns:
            # Grab the column
            values = dp[input_grounding[col]]
            # Remove it from the dp
            dp.remove_column(input_grounding[col])
            # Add again with the right column name
            dp.add_column(col, values)

        for col in self.output_schema.columns:
            # Grab the column
            values = dp[output_grounding[col]]
            # Remove it from the dp
            dp.remove_column(output_grounding[col])
            # Add again with the right column name
            dp.add_column(col, values)

        return dp

        # # Construct a map_fn that remaps the dataset schema
        # def _map_fn(example):
        #     return tz.merge(
        #         {k: example[input_grounding[k]] for k in self.input_schema.features},
        #         {k: example[output_grounding[k]] for k in self.output_schema.features
        #         },
        #     )
        #
        # return dataset.map(
        #     _map_fn,
        #     remove_columns=list(reversed_input_grounding.keys())
        #     + list(reversed_output_grounding.keys()),
        # )

    def classification(self):
        # TODO(karan): improve the schema inference
        # Check that the only output is a ClassLabel output
        if len(self._output_schema) == 1 and isinstance(
            self._output_schema.features[self._output_schema.columns[0]], ClassLabel
        ):
            return True
        return False


class Generic(Task):
    def __init__(self):
        super(Generic, self).__init__(
            identifier="Generic",
            input_schema=None,
            output_schema=None,
            metrics=None,
        )


class SentimentClassification(Task):
    def __init__(self, identifier, input_schema, output_schema, *args, **kwargs):
        super(SentimentClassification, self).__init__(
            identifier=identifier,
            input_schema=input_schema,
            output_schema=output_schema,
            metrics=[
                "accuracy",
                "f1",
                "class_dist",
                "pred_dist"
                # TODO(karan): calibration, other metrics
            ],
            *args,
            **kwargs,
        )


class BinarySentiment(SentimentClassification):
    def __init__(self):
        super(BinarySentiment, self).__init__(
            num_classes=2,
            input_schema=Schema(
                features=OrderedDict(
                    [
                        ("text", Value(dtype="string")),
                    ]
                ),
                grounding_candidates={
                    "text": {"text", "sentence"},
                },
            ),
            output_schema=Schema(
                features=OrderedDict(
                    [
                        ("label", ClassLabel(names=["negative", "positive"])),
                    ]
                ),
                grounding_candidates={
                    "label": {"label"},
                },
            ),
            identifier=self.__class__.__name__,
        )

    @classmethod
    def list_datasets(cls):
        return [
            "imdb",
        ]


class Summarization(Task):
    def __init__(self):
        super(Summarization, self).__init__(
            identifier=self.__class__.__name__,
            input_schema=Schema(
                features=OrderedDict([("text", Value(dtype="string"))]),
                grounding_candidates={
                    "text": {"article", "document"},
                },
            ),
            output_schema=Schema(
                features=OrderedDict([("summary", Value(dtype="string"))]),
                grounding_candidates={
                    "summary": {"highlights", "summary"},
                },
            ),
            metrics=[
                # blah,
                # TODO(karan): calibration, other metrics
                "rouge1",
                "rouge2",
                "rougeLsum",
            ],
        )

    @classmethod
    def list_datasets(cls):
        return [
            "cnn_dailymail",
        ]


class NaturalLanguageInference(Task):
    def __init__(self, identifier, input_schema, output_schema, *args, **kwargs):
        super(NaturalLanguageInference, self).__init__(
            identifier=identifier,
            input_schema=input_schema,
            output_schema=output_schema,
            metrics=[
                "accuracy",
                "f1_micro",
                "f1_macro",
                "class_dist",
                "pred_dist"
                # TODO(karan): calibration, other metrics
            ],
            *args,
            **kwargs,
        )


class BinaryNaturalLanguageInference(NaturalLanguageInference):
    def __init__(self):
        super(BinaryNaturalLanguageInference, self).__init__(
            num_classes=2,
            input_schema=Schema(
                features=OrderedDict(
                    [
                        ("premise", Value(dtype="string")),
                        ("hypothesis", Value(dtype="string")),
                    ]
                ),
                grounding_candidates={
                    "premise": {"premise", "sentence1"},
                    "hypothesis": {"hypothesis", "sentence2"},
                },
            ),
            output_schema=Schema(
                features=OrderedDict(
                    [
                        ("label", ClassLabel(names=["entailment", "non entailment"])),
                    ]
                ),
                grounding_candidates={
                    "label": {"label"},
                },
            ),
            identifier=self.__class__.__name__,
        )

    @classmethod
    def list_datasets(cls):
        return []


class TernaryNaturalLanguageInference(NaturalLanguageInference):
    def __init__(self):
        super(TernaryNaturalLanguageInference, self).__init__(
            num_classes=3,
            input_schema=Schema(
                features=OrderedDict(
                    [
                        ("premise", Value(dtype="string")),
                        ("hypothesis", Value(dtype="string")),
                    ]
                ),
                grounding_candidates={
                    "premise": {"premise", "sentence1"},
                    "hypothesis": {"hypothesis", "sentence2"},
                },
            ),
            output_schema=Schema(
                features=OrderedDict(
                    [
                        (
                            "label",
                            ClassLabel(
                                names=["entailment", "neutral", "contradiction"]
                            ),
                        ),
                    ]
                ),
                grounding_candidates={
                    "label": {"label"},
                },
            ),
            identifier=self.__class__.__name__,
        )

    def datasets(self):
        return {
            "snli",
        }


class QuestionAnswering(Task):
    def __init__(self, identifier, input_schema, output_schema, *args, **kwargs):
        super(QuestionAnswering, self).__init__(
            identifier=identifier,
            input_schema=input_schema,
            output_schema=output_schema,
            metrics=[
                "em",
                "f1",
                # TODO(karan): calibration, other metrics
            ],
            *args,
            **kwargs,
        )


class ExtractiveQuestionAnswering(Task):
    def __init__(self):
        super(ExtractiveQuestionAnswering, self).__init__(
            input_schema=Schema(
                features=OrderedDict(
                    [
                        ("context", Value(dtype="string")),
                        ("question", Value(dtype="string")),
                    ]
                ),
                grounding_candidates={
                    "context": {"context"},
                    "question": {"question"},
                },
            ),
            output_schema=Schema(
                features=OrderedDict(
                    [
                        (
                            "answers",
                            Sequence(
                                feature={
                                    "text": Value(dtype="string", id=None),
                                    "answer_start": Value(dtype="int32", id=None),
                                },
                                length=-1,
                            ),
                        ),
                    ]
                ),
                grounding_candidates={
                    "answers": {
                        "answers",
                    },
                },
            ),
            metrics=[
                "em",
                "f1",
            ],
            identifier=self.__class__.__name__,
        )



"""Spacy Operation."""
from __future__ import annotations

from typing import List

import torch
from meerkat import SpacyCell
from meerkat.tools.lazy_loader import LazyLoader

from robustnessgym.core.operation import Operation
from robustnessgym.core.slice import SliceDataPanel as DataPanel

spacy = LazyLoader("spacy", warning="Please `pip install spacy`.")
spacy_tokens = LazyLoader("spacy.tokens")
nc = LazyLoader(
    "neuralcoref",
    error="Can't import neuralcoref. Please install neuralcoref using:\n"
    "git clone https://github.com/huggingface/neuralcoref.git\n"
    "cd neuralcoref\n"
    "pip install -r requirements.txt\n"
    "pip install -e .",
)


class SpacyOp(Operation):
    """Operation that runs the Spacy pipeline."""

    def __init__(
        self,
        lang: str = "en_core_web_sm",
        nlp: spacy.language.Language = None,
        neuralcoref: bool = False,
        device: str = None,
    ):
        # Set all the parameters
        self.lang = lang
        self.neuralcoref = neuralcoref
        self._prebuilt = True

        # Set the device
        self._on_gpu = False
        if device and (device == "gpu" or device.startswith("cuda")):
            spacy.prefer_gpu(
                gpu_id=0 if ":" not in device else int(device.split(":")[1])
            )
            # Spacy sets the default torch float Tensor to torch.cuda.FloatTensor,
            # which causes other GPU operations to crash.
            torch.set_default_tensor_type("torch.FloatTensor")
            self._on_gpu = True

        # Load up the Spacy module
        self.nlp = nlp
        if not nlp:
            self.nlp = spacy.load(lang)
            self._prebuilt = False

        # Add neuralcoref
        if self.neuralcoref:
            nc.add_to_pipe(self.nlp)

        if not nlp:
            super(SpacyOp, self).__init__(
                lang=lang,
                neuralcoref=neuralcoref,
            )
        else:
            super(SpacyOp, self).__init__(
                lang=nlp.lang,
                # No need to pass in neuralcoref separately, it's already in the
                # pipeline if neuralcoref=True
                pipeline=nlp.pipe_names,
            )

    def process_batch(
        self,
        dp: DataPanel,
        columns: List[str],
        **kwargs,
    ) -> tuple:
        """Process text examples by running them through the Spacy pipeline.

        Args:
            dp (DataPanel): DataPanel
            columns (list): list of columns
            **kwargs: optional keyword arguments

        Returns:
            Tuple with single output, a list of SpacyCell objects.
        """
        if self._on_gpu:
            # Adjust the default Tensor type: this is instantaneous
            torch.set_default_tensor_type("torch.cuda.FloatTensor")

        # Apply Spacy's pipe method to process the examples
        cells = [SpacyCell(doc) for doc in list(self.nlp.pipe(dp[columns[0]]))]

        if self._on_gpu:
            # Reset the default Tensor type: this is instantaneous
            torch.set_default_tensor_type("torch.FloatTensor")

        return (cells,)


"""
if obj._.has("huggingface_neuralcoref"):
    # Create a helper function that turns a Span into a dictionary
    span_to_dict = lambda span: {
        "start": span.start,
        "end": span.end,
        "text": span.text,
    }

    # Create a helper function that converts a Cluster (output of
    # neuralcoref) into a dictionary
    cluster_to_dict = lambda cluster: {
        "i": cluster.i,
        "main": span_to_dict(cluster.main),
        "mentions": [span_to_dict(span) for span in cluster.mentions],
    }

    # Apply the helper functions to construct a dictionary for the
    # neuralcoref information
    neuralcoref_dict = {
        "neuralcoref": [
            cluster_to_dict(cluster) for cluster in obj._.coref_clusters
        ]
    }
"""



from typing import List

from robustnessgym.core.operation import Operation
from robustnessgym.core.slice import SliceDataPanel as DataPanel


class StripTextOp(Operation):
    def __init__(self):
        super(StripTextOp, self).__init__()

    def process_batch(
        self,
        dp: DataPanel,
        columns: List[str],
        **kwargs,
    ) -> tuple:
        # Clean up each text with a simple function and return the stripped text
        return (
            list(
                map(
                    lambda text: text.lower()
                    .replace(".", "")
                    .replace("?", "")
                    .replace("!", "")
                    .replace(",", ""),
                    dp[columns[0]],
                )
            ),
        )



"""Implementation of the Operation abstract base class."""
from __future__ import annotations

import pathlib
from abc import ABC
from collections import defaultdict
from functools import partial
from typing import Callable, Dict, List, Optional, Union

from meerkat import AbstractColumn
from meerkat.mixins.identifier import IdentifierMixin
from meerkat.tools.identifier import Identifier

from robustnessgym.core.slice import SliceDataPanel as DataPanel


def tuple_to_dict(keys: List[str]):
    def _tuple_to_dict(fn: callable):
        def _wrapper(*args, **kwargs):
            # Assume that if the output isn't a tuple,
            # it can be converted to a tuple of length 1
            output = fn(*args, **kwargs)
            if not isinstance(output, tuple):
                output = (output,)
            return dict(zip(keys, output))

        return _wrapper

    return _tuple_to_dict


def stow(
    dp: DataPanel,
    ops: Dict[Operation, List[List[str]]],
):
    """Apply Operations in sequence.

    Args:
        dp (DataPanel): DataPanel
        ops (Dict[Operation, List[List[str]]]):

    Returns:
        Updated DataPanel.
    """

    # Remove Operations whose outputs are already in the DataPanel
    for op, list_of_columns in list(ops.items()):
        indices_to_remove = []
        for i, columns in enumerate(list(list_of_columns)):
            if op.exists(dp):
                # Remove the columns at index i
                indices_to_remove.append(i)

        # Remove input columns for which the Operation was previously executed
        for index in sorted(indices_to_remove, reverse=True):
            columns = ops[op].pop(index)
            print(f"skipped: {op.identifier} -> {columns}", flush=True)

        # Remove the op entirely if list_of_columns is now empty
        if not ops[op]:
            ops.pop(op)

    # Run the remaining Operations
    for op, list_of_columns in ops.items():
        for columns in list_of_columns:
            dp = op(dp, columns=columns)

    return dp


def lookup(
    dp: DataPanel,
    op: Union[type, Operation],
    columns: List[str],
    output_name: str = None,
) -> AbstractColumn:
    """Retrieve the outputs of an Operation from a DataPanel.

    Args:
        dp (DataPanel): DataPanel
        op (Union[type, Operation]): subclass of Operation, or Operation object
        columns (List[str]): list of input columns that Operation was applied to
        output_name (Optional[str]): for an Operation with `num_outputs` > 1,
            the name of the output column to lookup

    Returns:
        Output columns of the Operation from the DataPanel.
    """
    # Operation identifier that should be retrieved
    if isinstance(op, Operation):
        op_name = str(op.identifier.name)
    else:
        op_name = str(Identifier(op.__name__))

    # Identifiers for all columns in the DataPanel, grouped without input columns
    # for Operation identifiers.
    column_identifiers = defaultdict(list)
    for col in dp.columns:
        identifier = Identifier.parse(col)
        column_identifiers[identifier.without("columns")].append(identifier)

    # Search for the column group that best matches the Operation identifier
    best_match, best_distance = None, 100000000
    for identifier in column_identifiers:
        # The prefix to match
        prefix = str(identifier)

        # Pick the key that best matches the cls name or instance identifier
        if (
            prefix.startswith(op_name)
            and len(
                prefix.replace(op_name, "").replace(
                    "" if output_name is None else output_name, ""
                )
            )
            < best_distance
        ):
            best_match = identifier
            best_distance = len(
                prefix.replace(op_name, "").replace(
                    "" if output_name is None else output_name, ""
                )
            )

    # Get the best matched column group
    identifier = best_match

    if identifier is None:
        raise AttributeError("Lookup failed.")

    return dp[str(identifier(columns=columns))]


class Operation(ABC, IdentifierMixin):
    """Abstract base class for operations in Robustness Gym."""

    # Path to a log directory
    logdir: pathlib.Path = pathlib.Path.home() / "robustnessgym/operations/"

    # Create a directory
    logdir.mkdir(parents=True, exist_ok=True)

    def __init__(
        self,
        identifier: Identifier = None,
        output_names: List[str] = None,
        process_batch_fn: Callable = None,
        prepare_batch_fn: Callable = None,
        **kwargs,
    ):
        super(Operation, self).__init__(
            identifier=identifier
            if identifier
            else Identifier(_name=self.__class__.__name__, **kwargs),
        )

        self._output_names = output_names

        if process_batch_fn:
            self.process_batch = process_batch_fn

        if prepare_batch_fn:
            self.prepare_batch = prepare_batch_fn

    def __repr__(self):
        return str(self.identifier)

    @property
    def num_outputs(self) -> int:
        """Number of output columns created by the Operation."""
        return len(self.output_names) if self.output_names else 1

    @property
    def output_names(self) -> Optional[List[str]]:
        """Name of output columns created by the Operation."""
        return self._output_names

    @property
    def output_identifiers(self) -> List[Identifier]:
        if self.output_names:
            return [self.identifier(output=name) for name in self.output_names]
        return [self.identifier]

    @property
    def output_columns(self) -> List[str]:
        return [str(identifier) for identifier in self.output_identifiers]

    @property
    def identifier(self) -> Identifier:
        """Name of the Operation."""
        return self._identifier

    @classmethod
    def exists(cls, dp: DataPanel) -> bool:
        """Check if the outputs of the Operation are in `dp`.

        Args:
            dp: DataPanel

        Returns:
            bool: True if `dp` contains a column produced by `Operation`,
                False otherwise
        """
        # TODO: update this to use `Operation.outputs`
        return any([key.startswith(cls.__name__) for key in dp.keys()])

    def prepare_batch(
        self,
        dp: DataPanel,
        columns: List[str],
        *args,
        **kwargs,
    ) -> None:
        """Preparation applied to the DataPanel `dp`.

        This is provided as a convenience function that can be called by
        `self.prepare`.

        Args:
            dp (DataPanel): DataPanel
            columns (list): list of columns
            *args: optional positional arguments
            **kwargs: optional keyword arguments
        """
        raise NotImplementedError("Implement `prepare_batch`.")

    def prepare(
        self,
        dp: DataPanel,
        columns: List[str],
        batch_size: int = 32,
        *args,
        **kwargs,
    ) -> None:
        """Preparation that is applied before the Operation is applied.

        Many Operations require a full pass over the DataPanel to precompute some
        variables before the core operation can actually be applied e.g. to create a
        Bag-of-Words representation, constructing a vocabulary to keep only
        tokens that are frequently seen across the DataPanel.

        Args:
            dp (DataPanel): DataPanel
            columns (list): list of columns
            batch_size (int): batch size for `dp.map(...)`
            *args: optional positional arguments
            **kwargs: optional keyword arguments
        """

        try:
            dp.map(
                function=partial(self.prepare_batch, columns=columns, *args, **kwargs),
                input_columns=columns,
                is_batched_fn=True,
                batch_size=batch_size,
                *args,
                **kwargs,
            )
        except NotImplementedError:
            return

    def process_batch(
        self,
        dp: DataPanel,
        columns: List[str],
        **kwargs,
    ) -> tuple:
        """The core functionality of the Operation.

        This is provided as a convenience function that can be called by
        `self.process`.

        Args:
            dp (DataPanel): DataPanel
            columns (list): list of columns
            **kwargs: optional keyword arguments

        Returns:
            Tuple of outputs, where each output is a a sequence of values. The expected
                order of the outputs is the same as the order of identifiers in
                `self.outputs`.
        """
        return NotImplemented

    def process(
        self,
        dp: DataPanel,
        columns: List[str],
        batch_size: int = 32,
        *args,
        **kwargs,
    ) -> DataPanel:
        """Apply the Operation to a DataPanel.

        Args:
            dp (DataPanel): DataPanel
            columns (list): list of columns
            batch_size (int): batch size for `dp.update(...)`
            *args: optional positional arguments
            **kwargs: optional keyword arguments
        """

        return dp.update(
            tuple_to_dict(
                keys=[str(ident(columns=columns)) for ident in self.output_identifiers]
            )(partial(self.process_batch, columns=columns, *args, **kwargs)),
            batch_size=batch_size,
            is_batched_fn=True,
            *args,
            **kwargs,
        )

    def __call__(
        self,
        dp: DataPanel,
        columns: List[str],
        batch_size: int = 32,
        **kwargs,
    ) -> DataPanel:
        """Apply the Operation to a DataPanel.

        Args:
            dp (DataPanel): DataPanel
            columns (list): list of columns
            batch_size (int):
            **kwargs: optional keyword arguments

        Returns:
            An updated DataPanel, with additional output columns produced by
              the Operation.
        """

        if isinstance(dp, DataPanel):
            assert len(set(columns) - set(dp.column_names)) == 0, (
                f"All `columns` ({columns}) must be present and visible in `dp` ("
                f"{list(dp.column_names)})."
            )

            if self.exists(dp):
                return dp

            # Prepare to apply the Operation to the DataPanel
            self.prepare(
                dp=dp,
                columns=columns,
                batch_size=batch_size,
                **kwargs,
            )

            # Apply the Operation to the DataPanel
            dp = self.process(
                dp=dp,
                columns=columns,
                batch_size=batch_size,
                **kwargs,
            )

            return dp

        else:
            return self(
                dp=DataPanel(dp),
                columns=columns,
                batch_size=batch_size,
                **kwargs,
            )



from __future__ import annotations

import json
from json import JSONDecodeError
from typing import Callable, Dict, List, Optional, Union

from meerkat import AbstractColumn, DataPanel

from robustnessgym.core.constants import CURATION, GENERIC, SUBPOPULATION
from robustnessgym.core.identifier import Id, Identifier


class SliceMixin:
    """Slice class in Robustness Gym."""

    def __init__(self):
        # A slice has a lineage
        if self.identifier is None:
            self.lineage = []
        else:
            self.lineage = [(str(self.__class__.__name__), self.identifier)]

        # Set the category of the slice: defaults to 'curated'
        self.category = CURATION

    def add_to_lineage(self, category, identifier, columns=None):
        """Append to the lineage."""
        # TODO (karan): add Identifier directly
        if columns:
            self.lineage.append((category, identifier, columns))
        else:
            self.lineage.append((category, identifier))

        # Update the identifier
        self._lineage_to_identifier()

    def _add_op_to_lineage(self):
        if self.node.last_parent is not None:
            opnode, indices = self.node.last_parent
            try:
                fn = opnode.captured_args["function"]
            except KeyError:
                return

            if opnode.ref().__name__ == "filter":
                self.add_to_lineage(
                    SUBPOPULATION.capitalize(),
                    Id("Function", name=fn.__name__, mem=hex(id(fn))),
                    [],
                )
                self.category = SUBPOPULATION
            else:
                self.add_to_lineage(
                    GENERIC.capitalize(),
                    Id("Function", name=fn.__name__, mem=hex(id(fn))),
                    [],
                )

    def _lineage_to_identifier(self):
        """Synchronize to the current lineage by reassigning to
        `self._identifier`."""
        short_lineage = []
        for entry in self.lineage:
            if len(entry) == 3:
                try:
                    columns = json.loads(entry[2])
                except JSONDecodeError:
                    columns = entry[2]
                short_lineage.append(str(entry[1]) + " @ " + str(columns))
            else:
                short_lineage.append(str(entry[1]))
        # Assign the new lineage to the identifier
        self._identifier = Identifier(_name=" -> ".join(short_lineage))

    @property
    def identifier(self):
        """Slice identifier."""
        if self._identifier:
            return self._identifier
        if self.lineage:
            self._lineage_to_identifier()
            return self._identifier
        return None

    @identifier.setter
    def identifier(self, value):
        """Set the slice's identifier."""
        self._identifier = value

    @classmethod
    def _add_state_keys(cls) -> set:
        """List of attributes that describe the state of the object."""
        return {
            "lineage",
            "category",
        }


class SliceDataPanel(DataPanel, SliceMixin):
    def __init__(self, *args, **kwargs):
        super(SliceDataPanel, self).__init__(*args, **kwargs)
        SliceMixin.__init__(self)

    @classmethod
    def _state_keys(cls) -> set:
        state_keys = super(SliceDataPanel, cls)._state_keys()
        state_keys.union(cls._add_state_keys())
        return state_keys

    def update(
        self,
        function: Optional[Callable] = None,
        with_indices: bool = False,
        input_columns: Optional[Union[str, List[str]]] = None,
        is_batched_fn: bool = False,
        batch_size: Optional[int] = 1,
        remove_columns: Optional[List[str]] = None,
        num_workers: int = 0,
        materialize: bool = True,
        pbar: bool = False,
        **kwargs,
    ) -> SliceDataPanel:
        dp = super(SliceDataPanel, self).update(
            function=function,
            with_indices=with_indices,
            input_columns=input_columns,
            is_batched_fn=is_batched_fn,
            batch_size=batch_size,
            remove_columns=remove_columns,
            num_workers=num_workers,
            materialize=materialize,
            pbar=pbar,
            **kwargs,
        )
        if isinstance(dp, SliceDataPanel):
            dp._add_op_to_lineage()

        return dp

    def filter(
        self,
        function: Optional[Callable] = None,
        with_indices=False,
        input_columns: Optional[Union[str, List[str]]] = None,
        is_batched_fn: bool = False,
        batch_size: Optional[int] = 1,
        drop_last_batch: bool = False,
        num_workers: int = 0,
        materialize: bool = True,
        pbar: bool = False,
        **kwargs,
    ) -> Optional[SliceDataPanel]:
        dp = super(SliceDataPanel, self).filter(
            function=function,
            with_indices=with_indices,
            input_columns=input_columns,
            is_batched_fn=is_batched_fn,
            batch_size=batch_size,
            drop_last_batch=drop_last_batch,
            num_workers=num_workers,
            materialize=materialize,
            pbar=pbar,
            **kwargs,
        )
        if isinstance(dp, SliceDataPanel):
            dp._add_op_to_lineage()
        return dp

    def map(
        self,
        function: Optional[Callable] = None,
        with_indices: bool = False,
        input_columns: Optional[Union[str, List[str]]] = None,
        is_batched_fn: bool = False,
        batch_size: Optional[int] = 1,
        drop_last_batch: bool = False,
        num_workers: int = 0,
        output_type: type = None,
        mmap: bool = False,
        materialize: bool = True,
        pbar: bool = False,
        **kwargs,
    ) -> Optional[Union[Dict, List, AbstractColumn]]:
        dp = super(SliceDataPanel, self).map(
            function=function,
            with_indices=with_indices,
            input_columns=input_columns,
            is_batched_fn=is_batched_fn,
            batch_size=batch_size,
            drop_last_batch=drop_last_batch,
            num_workers=num_workers,
            output_type=output_type,
            mmap=mmap,
            materialize=materialize,
            pbar=pbar,
            **kwargs,
        )
        if isinstance(dp, SliceDataPanel):
            dp._add_op_to_lineage()
        return dp



from __future__ import annotations

import inspect
import json
import logging
import pathlib
from typing import Callable, Collection, Dict, List, Union

import dill
import pandas as pd
from fuzzywuzzy import process
from tqdm import tqdm

from robustnessgym.core.constants import (
    ATTACK,
    AUGMENTATION,
    CURATION,
    GENERIC,
    SUBPOPULATION,
)
from robustnessgym.core.slice import SliceDataPanel as DataPanel
from robustnessgym.core.tools import persistent_hash
from robustnessgym.core.version import SemanticVersionerMixin
from robustnessgym.report.report import NumericColumn, Report, ReportColumn, ScoreColumn
from robustnessgym.slicebuilders.slicebuilder import SliceBuilder

logger = logging.getLogger(__name__)


class ReportableMixin:
    def _shared_aggregators(self, models: List[str]):
        """Find aggregators that are shared by multiple models in the Bench."""

        common_aggregators = []
        # Iterate over all the models
        for model in models:
            assert model in self.metrics, f"Model {model} does not exist."
            common_aggregators.append(set([agg for agg in self.aggregators[model]]))

        # Find the common aggregators by taking an intersection
        return set.intersection(*[set(e) for e in common_aggregators])

    def create_report(
        self,
        models: List[str] = None,
        aggregator_columns: Dict[str, ReportColumn] = None,
    ) -> Report:
        """Generate a report for models in the bench.

        Args:
            models (List[str]): names of one or more models that are in the devbench.
            aggregator_columns (Dict[str, (ReportColumn, dict)]):
                dict mapping aggregator names to a tuple.

                The first entry of the tuple is the ReportColumn that should be
                used for visualization. The second entry is a dict of kwargs that
                will be passed to the ReportColumn using
                `ReportColumn.__init__(..., **kwargs)`.

                For instance,
                >>> devbench.create_report(
                >>>     models=['BERT'],
                >>>     aggregator_columns={
                >>>         'accuracy': (ScoreColumn, {'min_val': 0.3})
                >>>     }
                >>> )

                By default, aggregators will be displayed as a ScoreColumn
                with `min_val=0`, `max_val=1` and `is_0_to_1=True`.

        Returns:
            a Report, summarizing the performance of the models.
        """

        if len(self.slices) == 0:
            raise ValueError("No slices found in Bench. Cannot create report.")

        if models is not None:
            for model in models:
                assert model in self.metrics, f"Model {model} not found."
        else:
            # Use all the models that are available
            models = list(self.metrics.keys())

        # Set identifiers to be human readable
        self._human_readable_identifiers()

        # Get the list of aggregators that are shared by `models`
        shared_aggregators = list(self._shared_aggregators(models))

        # Populate columns
        columns = []
        for model in models:
            for aggregator in shared_aggregators:
                if aggregator_columns and aggregator in aggregator_columns:
                    column_type, column_kwargs = aggregator_columns[aggregator]
                else:
                    column_type = ScoreColumn
                    column_kwargs = dict(min_val=0, max_val=1, is_0_to_1=True)
                columns.append(column_type(f"{model}-{aggregator}", **column_kwargs))
        columns.append(NumericColumn("Size"))

        category_names = {
            GENERIC: "Slice",
            SUBPOPULATION: "SubPop",
            ATTACK: "Attack",
            AUGMENTATION: "Augment",
            CURATION: "Eval",
        }

        # Populate data
        data = []
        for sl in self.slices:
            slice_name = self.ident_mapping[sl.identifier]
            slice_size = len(sl)
            slice_category = category_names.get(sl.category, sl.category.capitalize())

            row = [slice_category, slice_name]

            for model in models:
                model_metrics = self.metrics[model]
                if sl.identifier not in model_metrics:
                    continue
                slice_metrics = model_metrics[sl.identifier]
                for agg in shared_aggregators:
                    row.append(slice_metrics[agg])

            row.append(slice_size)
            data.append(row)

        df = pd.DataFrame(data)

        report = Report(
            data=df,
            columns=columns,
        )
        report.sort(
            category_order=dict(
                (cat, i)
                for i, cat in enumerate(
                    [SUBPOPULATION, AUGMENTATION, CURATION, ATTACK, GENERIC]
                )
            )
        )
        return report

    def _human_readable_identifiers(self):
        # Temporary function to generate human readable names
        groups = {}
        for ident in self._slice_identifiers:
            if "->" in str(ident):
                builder_ident = str(ident).split(" -> ")[-1]
                try:
                    builder_ident, cols = builder_ident.split(" @ ")
                except ValueError:
                    cols = ""
                name = builder_ident.split("(")[0]
                if name not in groups:
                    groups[name] = set()
                groups[name].add((builder_ident, cols))

        group_info = {}
        for key, group in groups.items():
            if len(group) == 1:
                group_info[key] = "name"
            else:
                only_single_column = len(set([t[1] for t in group])) == 1
                if only_single_column:
                    group_info[key] = "builder_ident"
                else:
                    group_info[key] = "full"

        ident_mapping = {}
        for ident in self._slice_identifiers:
            if "->" in str(ident):
                builder_ident = str(ident).split(" -> ")[-1]
                try:
                    builder_ident, cols = builder_ident.split(" @ ")
                except ValueError:
                    cols = ""
                name = builder_ident.split("(")[0]

                if group_info[name] == "name":
                    new_ident = name
                elif group_info[name] == "builder_ident":
                    new_ident = builder_ident
                elif group_info[name] == "full":
                    new_ident = str(ident).split(" -> ")[-1]

                if new_ident.startswith("NlpAugTransformation"):
                    new_ident = new_ident.split("NlpAugTransformation(pipeline=[")[
                        1
                    ].split("])")[0]

            else:
                new_ident = str(ident).split("(")[0]

            ident_mapping[ident] = new_ident

        self.ident_mapping = ident_mapping


class DevBench(SemanticVersionerMixin, ReportableMixin):
    def __init__(self):
        super(DevBench, self).__init__()

        # The collection of slices
        self._slices = set()
        self._slice_identifiers = set()
        self._slice_table = {}

        # The devbench has aggregators
        self.aggregators = {}

        # The devbench internally tracks metrics
        self.metrics = {}

    @property
    def slices(self):
        """Slices in the devbench."""
        return self._slices

    @property
    def models(self):
        """Models in the devbench."""
        return list(self.aggregators.keys())

    @property
    def summary(self):
        """Summary of the devbench."""
        return pd.DataFrame(self.metrics)

    def __repr__(self):
        return f"{self.__class__.__name__}(slices={len(self.slices)})"

    def __call__(self, slicebuilder: SliceBuilder, dp: DataPanel, columns: List[str]):
        return self.add_slices(slicebuilder(dp, columns)[0])

    def _digest(self) -> str:
        return json.dumps([str(sl) for sl in self.slices])

    def add_slices(
        self,
        slices: Union[DataPanel, Collection[DataPanel]],
        overwrite: bool = False,
    ) -> None:
        """Add slices to the DevBench.

        Args:
            slices (Union[DataPanel, Collection[DataPanel]]): a single DataPanel or
                a collection of DataPanels
            overwrite (bool): overwrite any slice in `slices` if it already exists
                in the DevBench
        """
        if isinstance(slices, DataPanel):
            slices = [slices]

        assert all(
            [isinstance(sl, DataPanel) for sl in slices]
        ), "All slices must be DataPanels."

        if not overwrite and any(
            [sl.identifier in self._slice_identifiers for sl in slices]
        ):
            logger.warning(
                "Some slices in `slices` already exist in the DevBench, "
                "pass `overwrite=True` to overwrite them."
            )

        for sl in slices:
            if overwrite or (
                sl.identifier not in self._slice_identifiers and len(sl) > 0
            ):
                # Add slices that aren't present and have non-zero length
                self._slices.add(sl)
                self._slice_identifiers.add(sl.identifier)
                self._slice_table[sl.identifier] = sl

        # Calculate all metrics
        self.calculate()

    def add_aggregators(
        self,
        aggregators: Dict[str, Dict[str, Callable]],
    ) -> None:
        """Add functions for aggregation, with a dictionary that maps a model
        name to aggregation functions.

        Args:
            aggregators (Dict[str, Dict[str, Callable]]): a dict with keys that
                are model names (str), and values that are also dicts.

                The value dicts map aggregator names (str) to
                aggregation functions that produce a value. Typically, the aggregator
                name will be a metric (e.g. "accuracy") of some kind.

                The aggregation function must be a single argument Callable that takes
                as input a DataPanel, and outputs a metric value. As an example,

                >>> def accuracy(dp: DataPanel) -> float:
                >>>    return (dp['pred'] == dp['label']).mean()

                The output type of the aggregation function can be arbitrary,
                e.g. metrics that are np.ndarray or strings are also supported.
        """

        # `aggregators`: model_name -> dict
        for model, aggs in aggregators.items():
            # `aggs`: aggregator_name -> Callable
            for agg_name, agg in aggs.items():
                assert isinstance(
                    agg, Callable
                ), f"Aggregators must be functions, but {agg} is not."
                arguments = inspect.getfullargspec(agg).args
                assert (
                    len(arguments) == 1
                ), f"Aggregators must be single argument but {agg} is not."

                # Store the aggregator
                if model not in self.aggregators:
                    self.aggregators[model] = {}
                self.aggregators[model][agg_name] = agg

        # Calculate all metrics
        self.calculate()

    def calculate(self) -> None:
        """Calculate all metrics that haven't been calculated yet.

        The DevBench internally tracks metrics for all models and aggregators that
        were previously added, across all added slices.

        This `calculate` method is triggered automatically when a new slice
        or a new model and aggregator are added to the DevBench.
        """

        # `self.aggregators`: model_name -> aggregator_name
        for model in self.aggregators:

            # Add the model to the metrics dict
            if model not in self.metrics:
                self.metrics[model] = {}

            for sl in self.slices:
                # Add the slice to the model's metrics dict
                if str(sl.identifier) not in self.metrics[model]:
                    self.metrics[model][str(sl.identifier)] = {}

                for agg_name, aggregator in self.aggregators[model].items():
                    if agg_name not in self.metrics[model][str(sl.identifier)]:
                        # Calculate the aggregated value
                        self.metrics[model][str(sl.identifier)][agg_name] = aggregator(
                            sl
                        )

    def search(self, keyword: str, limit: int = 3) -> List[DataPanel]:
        """Fuzzy search to find a slice in the devbench. Returns slices that
        have identifiers that best match the searched keyword.

        Args:
            keyword (str): search phrase. The keyword will be used to match slice
                identifiers in the devbench.
            limit (int): number of results to return. Defaults to 3.

        Returns:
            A list of slice DataPanels.
        """
        return [
            self._slice_table[t[0]]
            for t in process.extract(keyword, self._slice_identifiers, limit=limit)
        ]

    def save(self, path: str) -> None:
        """Save the devbench to disk.

        This will save all slices in the devbench to disk, as well as
        metrics and aggregators associated with this devbench.

        Args:
            path: string path to the save directory e.g. "./my_analysis".
                A `.devbench` extension is added, so the devbench will be stored at
                "./my_analysis.devbench".

                To load the devbench back in, use `DevBench.load("./my_analysis")` or
                `DevBench.load("./my_analysis.devbench")`.
        """
        # Path to the save directory
        savedir = pathlib.Path(path)
        savedir = savedir.with_suffix(".devbench")

        # Create a directory inside savedir for the slices
        (savedir / "slices").mkdir(parents=True, exist_ok=True)

        # Save all the slices
        pbar = tqdm(self.slices)
        for sl in pbar:
            pbar.set_description(f"Saving slice {str(sl.identifier)[:100]}...")
            sl.write(str(savedir / "slices" / str(persistent_hash(str(sl.identifier)))))

        # Save metrics
        dill.dump(self.metrics, open(str(savedir / "metrics.dill"), "wb"))

        # Save aggregators
        dill.dump(self.aggregators, open(str(savedir / "aggregators.dill"), "wb"))

        # Save version info
        with open(str(savedir / "version.dill"), "wb") as f:
            f.write(self._dumps_version())

    @classmethod
    def available(cls, path: str) -> List[str]:
        """Check what devbenches are available in a directory.

        Args:
            path (str): directory path.
                The devbenches available inside this directory will be returned.

        Returns:
            names of available devbenches in a list.
        """

        # Path to the save directory
        savedir = pathlib.Path(path)

        devbenches = []
        for maybe_devbench in savedir.glob("*"):

            # DevBench is saved to a directory
            if maybe_devbench.is_dir():
                if (maybe_devbench / "metadata.dill").exists():  # TODO: deprecate
                    devbenches.append(maybe_devbench.name)
                elif maybe_devbench.suffix == ".devbench":
                    devbenches.append(maybe_devbench.name)
                else:
                    continue

        return devbenches

    @classmethod
    def load(cls, path: str) -> DevBench:
        """Load a devbench from disk.

        Args:
            path (str): path to the devbench directory. The devbench directory must
                have the `.devbench` extension.

        Returns:
            a DevBench
        """

        # Path to the save directory
        savedir = pathlib.Path(path)

        # Load all the slices
        slices = []
        for sl_path in tqdm(list((savedir / "slices").glob("*"))):
            try:
                slices.append(DataPanel.read(str(sl_path)))
            except FileNotFoundError:
                continue

        # Load metrics
        metrics = dill.load(open(str(savedir / "metrics.dill"), "rb"))

        # Load metrics
        aggregators = dill.load(open(str(savedir / "aggregators.dill"), "rb"))

        # Create the devbench
        devbench = cls()

        # Set previously stored metrics
        devbench.metrics = metrics

        # Set previously stored aggregators
        devbench.aggregators = aggregators

        # Set the slices
        devbench.add_slices(slices)

        # Load version info
        with open(str(savedir / "version.dill"), "rb") as f:
            devbench._loads_version(f.read())

        return devbench



import itertools

import pandas as pd
from meerkat import DataPanel
from meerkat.columns.prediction_column import ClassificationOutputColumn

from robustnessgym.core.model import Model
from robustnessgym.tasks.task import ExtractiveQuestionAnswering


def run(dataset_info, model_name, debug=False):
    def fn(dp):
        # TODO(karan): Uncomment & add correct code here.
        # batch = {"col1": dp["context"].data, "col2": dp["question"].data}
        # input_cols = ["context", "question"]
        #
        # input_batch = model.encode_batch(batch, input_cols)
        # import cytoolz as tz
        # import torch
        # input_batch = tz.valmap(
        #     lambda v: torch.tensor(v).to(device=model.device),
        #     input_batch
        # )
        #
        # with torch.no_grad():
        #     outputs = model.model(**input_batch)
        #
        # outputs = (torch.argmax(outputs.start_logits).item(),
        #            torch.argmax(outputs.end_logits).item())
        #
        # token_id = input_batch['input_ids'][0][outputs[0]: outputs[1]]
        # answer = model.tokenizer.decode(token_id)
        # return answer
        pass

    dataset_name, split = dataset_info
    if isinstance(dataset_name, str):
        dataset_name = (dataset_name,)
    dp = DataPanel.load_huggingface(*dataset_name, split=split)
    if debug:
        dp = dp[:10]

    model = Model.huggingface(
        model_name,
        task=ExtractiveQuestionAnswering(),
        is_classifier=False,
    )

    out = dp.map(
        fn,
        batch_size=5,
        is_batched_fn=True,
        # output_type=ClassificationOutputColumn,
        pbar=True,
    )

    out = ClassificationOutputColumn(logits=out)

    del model

    dirname = f"./{'_'.join(dataset_name)}-{model_name}"
    out.write(dirname)

    # TO LOAD:
    # ClassificationOutputColumn.read(dirname)

    return out


DATASETS = [
    ("squad", "validation"),
]

# 'squad' 'validation'

MODELS = [
    "distilbert-base-cased-distilled-squad",
]

ds = DATASETS[0]
model = MODELS[0]

outs = []
for ds, model in itertools.product(DATASETS, MODELS):
    try:
        run(ds, model, debug=True)
        outs.append(
            {"dataset": str(ds), "model": model, "status": "Success", "Reason": None}
        )
    except Exception as e:
        print(e)
        outs.append(
            {"dataset": str(ds), "model": model, "status": "Failed", "Reason": str(e)}
        )

success = pd.DataFrame(outs)



from typing import Dict, List, Optional

from meerkat.tools.lazy_loader import LazyLoader
from torch import cuda

from robustnessgym.core.operation import Operation
from robustnessgym.core.slice import SliceDataPanel as DataPanel

bootleg_annotator = LazyLoader(
    "bootleg.end2end.bootleg_annotator",
    warning="Bootleg not available for import. Please see "
    "https://bootleg.readthedocs.io/en/latest/gettingstarted/install.html "
    "for help getting started.",
)


class BootlegAnnotatorOp(Operation):
    def __init__(
        self,
        config: Optional[Dict] = None,
        device: Optional[int] = None,
        cand_map: Optional[int] = None,
        threshold: Optional[float] = 0.5,
        cache_dir: Optional[str] = None,
        model_name: Optional[str] = None,
    ):

        # Pass threshold to Operation to use as identifier
        super(BootlegAnnotatorOp, self).__init__(threshold=threshold)

        # Set the device
        if not device:
            device = "cuda" if cuda.is_available() else "cpu"

        # Create the annotator
        self.annotator = bootleg_annotator.BootlegAnnotator(
            config=config,
            device=device,
            cand_map=cand_map,
            threshold=threshold,
            cache_dir=cache_dir,
            model_name=model_name,
        )

    def process_batch(
        self,
        dp: DataPanel,
        columns: List[str],
        **kwargs,
    ) -> tuple:
        """Annotate text samples using a Bootleg Annotator.

        Args:
            dp (DataPanel): DataPanel
            columns (list): list of columns
            **kwargs: optional keyword arguments

        Returns:
            Tuple with single output, a list of Bootleg annotations.
        """
        return ([self.annotator.label_mentions(text) for text in dp[columns[0]]],)



from __future__ import annotations

import pathlib

# from functools import partial
from itertools import compress
from typing import Callable, Dict, List, Optional, Sequence, Tuple, Union

import cytoolz as tz
import meerkat
import numpy as np
from meerkat.provenance import capture_provenance

from robustnessgym.core.constants import (
    ATTACK,
    CURATION,
    GENERIC,
    SLICEBUILDERS,
    SUBPOPULATION,
    TRANSFORMATION,
)
from robustnessgym.core.identifier import Identifier
from robustnessgym.core.slice import SliceDataPanel as DataPanel
from robustnessgym.core.storage import StorageMixin
from robustnessgym.core.tools import strings_as_json


class SliceBuilder(StorageMixin):
    """Base class for builders that output slices."""

    # Path to a log directory
    logdir: pathlib.Path = pathlib.Path.home() / "robustnessgym" / SLICEBUILDERS

    # Create the log directory
    logdir.mkdir(parents=True, exist_ok=True)

    CATEGORIES = [
        GENERIC,
        SUBPOPULATION,
        ATTACK,
        TRANSFORMATION,
        CURATION,
    ]

    def __init__(
        self,
        identifiers: List[Identifier],
        category: str = GENERIC,
        apply_fn: Callable = None,
        *args,
        **kwargs,
    ):

        super(SliceBuilder, self).__init__(*args, **kwargs)

        # The SliceBuilder belongs to a category
        assert (
            category in self.CATEGORIES
        ), f"argument category must be one of {self.CATEGORIES}"
        self.category = category

        # Each identifier corresponds to a Slice generated by this SliceBuilder
        self.identifiers = identifiers

        # Keep track of the Operation dependencies
        self.prerequisites = (
            set() if "prerequisites" not in kwargs else kwargs["prerequisites"]
        )

        if apply_fn:
            # Assign to the method
            self.apply = apply_fn

    def __repr__(self):
        return (
            f"{self.category}[{self.__class__.__name__}(num_slices={self.num_slices})]"
        )

    @property
    def num_slices(self):
        return len(self.identifiers)

    def __getitem__(self, item: int):
        return self.identifiers[item]

    def __iter__(self):
        yield from self.identifiers

    @capture_provenance(capture_args=["self", "dp", "columns", "batch_size"])
    def __call__(
        self,
        dp: Optional[DataPanel],
        columns: List[str],
        batch_size: int = 100,
        num_proc: int = None,
        *args,
        **kwargs,
    ):

        # Check that prerequisites are satisfied
        self.prerequisites_handler(dp)

        if isinstance(dp, DataPanel):
            # Prepare the data
            self.prepare_dataset(
                dp=dp,
                columns=columns,
                batch_size=batch_size,
                *args,
                **kwargs,
            )

            # Slice a dataset
            slices, slice_membership = self.process_dataset(
                dp=dp,
                columns=columns,
                num_proc=num_proc,
                *args,
                **kwargs,
            )

            return slices, slice_membership

        else:
            return self(
                dp=DataPanel(dp),
                columns=columns,
                *args,
                **kwargs,
            )

    def prerequisites_handler(
        self,
        dp: DataPanel,
    ):
        """Check if the DataPanel satisfies necessary prerequisites in order to
        run the SliceBuilder."""
        # Check if prerequisites are satisfied
        # TODO(karan): move to a method
        pending = {
            prerequisite
            for prerequisite in self.prerequisites
            if not prerequisite.available(dp)
        }

        # TODO(karan): Automatically run the pending pre-requisites
        if pending:
            raise RuntimeError(
                f"Cannot run SliceBuilder, prerequisites {pending} not satisfied."
            )

    def prepare_batch(
        self,
        batch: DataPanel,
        columns: List[str],
        *args,
        **kwargs,
    ) -> None:
        """Apply a preparation function to a batch. Use this to update
        attributes of `self`.

        Args:
            batch: batch of data
            columns: list of columns
            *args: optional additional arguments
            **kwargs: optional additional keyword arguments
        """
        raise NotImplementedError("Implement `prepare_batch`.")

    def _filter_prerequisite_columns(
        self,
        columns: List[str],
        all_columns: List[str],
    ) -> List[str]:
        # Simple filtering that doesn't use columns
        # TODO(karan): improve this by using `columns` to filter further
        return [
            col
            for col in all_columns
            if any(
                [
                    col.startswith(
                        prereq.__name__ if isinstance(prereq, type) else str(prereq)
                    )
                    for prereq in self.prerequisites
                ]
            )
        ]

    def prepare_dataset(
        self,
        dp: DataPanel,
        columns: List[str],
        batch_size: int = 32,
        *args,
        **kwargs,
    ) -> None:
        """Apply a preparation function to the data. Use this to update
        attributes of `self`.

        Args:
            dp: DataPanel
            columns: list of columns
            batch_size: batch size for preparation
            *args: optional additional arguments
            **kwargs: optional additional keyword arguments
        """
        # Set the data format
        with dp.format(
            columns + self._filter_prerequisite_columns(columns, dp.column_names)
        ):
            # Batch the dataset, and prepare each batch
            for batch in dp.batch(batch_size):
                try:
                    # Check if the `prepare_batch` function has been implemented
                    self.prepare_batch(
                        batch=batch,
                        columns=columns,
                        *args,
                        **kwargs,
                    )
                except NotImplementedError:
                    break

    def process_batch(
        self,
        dp: DataPanel,
        columns: List[str],
        *args,
        **kwargs,
    ) -> Tuple[Optional[List[DataPanel]], Optional[np.ndarray]]:
        """Apply a SliceBuilder to a batch of data.

        Args:
            dp: a DataPanel of data
            columns: list of columns
            *args: optional additional arguments
            **kwargs: optional additional keyword arguments

        Returns: tuple of
        (list of slices (as batches), matrix of (example, slice) membership))
        """
        return [dp], None

    def process_dataset(
        self,
        dp: DataPanel,
        columns: List[str],
        batch_size: int = 32,
        num_proc: int = None,
        *args,
        **kwargs,
    ) -> Tuple[List[DataPanel], np.ndarray]:
        """Apply a SliceBuilder to a dataset.

        Args:
            dp: DataPanel
            columns: list of columns
            batch_size: integer batch size
            num_proc: num processes for multiprocessing
            *args: optional additional arguments
            **kwargs: optional additional keyword arguments

        Returns: tuple of (DataPanel, list of Slices,
        matrix of (example, slice) membership)
        """
        # Create slices
        slices = [[DataPanel()] for _ in range(len(self.identifiers))]
        all_slice_memberships = []

        # Batch the dataset, and process each batch
        for batch in dp.batch(batch_size):
            # Process the batch
            sliced_batches, slice_memberships = self.process_batch(
                dp=batch,
                columns=columns,
                *args,
                **kwargs,
            )

            # Incrementally build the slices
            for sl, sl_batch in zip(slices, sliced_batches):
                sl.append(DataPanel(sl_batch))

            # Keep track of the slice memberships
            all_slice_memberships.append(slice_memberships)

        # Create a single slice label matrix
        slice_membership = np.concatenate(all_slice_memberships, axis=0)

        # Create a single DataPanel for each slice
        slices = [meerkat.concat(e[1:], axis=0) if len(e) > 1 else e[0] for e in slices]

        # TODO(karan): DataPanel doesn't support this
        for i, sl in enumerate(slices):
            # Set the Slice category using the SliceBuilder's category
            sl.category = self.category

            # Append the the lineage
            sl.add_to_lineage(
                category=str(self.category.capitalize()),
                identifier=self.identifiers[i],
                columns=strings_as_json(columns),
            )

        return slices, slice_membership

    def apply(self, *args, **kwargs):
        raise NotImplementedError("Must implement apply.")

    @classmethod
    def join(cls, *slicebuilders: SliceBuilder) -> Sequence[SliceBuilder]:
        """Join many SliceBuilders.

        By default, just returns them.
        """
        return slicebuilders

    @staticmethod
    def filter_batch_by_slice_membership(
        batch: Dict[str, List],
        slice_membership: np.ndarray,
    ) -> List[Dict[str, List]]:
        """Use a matrix of slice membership labels to select the subset of
        examples in each slice.

        Returns a list. Each element in the list corresponds to a single
        slice, and contains the subset of examples in 'batch' that lies
        in that slice.
        """
        return [
            tz.valmap(lambda v: list(compress(v, s)), batch) for s in slice_membership.T
        ]

    @classmethod
    def retrieve(
        cls,
        batch: DataPanel,
        columns: Union[List[str], List[List[str]]],
        proc_fns: Union[str, Callable, List[Union[str, Callable]]] = None,
        identifier: Union[str, Identifier] = None,
        reapply: bool = False,
        **kwargs,
    ) -> Optional[Union[DataPanel, List[DataPanel]]]:
        if not reapply:
            if "slices" not in batch:
                return None

            # Infer the most relevant key to retrieve if an identifier is not specified
            if not identifier:
                for ident_key in batch["slices"][0].keys():
                    # Pick the first key that matches the cls name
                    if ident_key.startswith(cls.__name__):
                        identifier = ident_key
                        break

            try:
                if isinstance(columns[0], str):
                    retrieval = {
                        strings_as_json(columns): [
                            cls.decode(cache[str(identifier)][strings_as_json(columns)])
                            for cache in batch["cache"]
                        ]
                    }
                else:
                    retrieval = {
                        strings_as_json(cols_): [
                            cls.decode(cache[str(identifier)][strings_as_json(cols_)])
                            for cache in batch["cache"]
                        ]
                        for cols_ in columns
                    }
            except KeyError:
                raise ValueError("Could not retrieve information for all keys.")

            # Check if the retrieved information needs to be processed
            if not proc_fns:
                return retrieval
            pass
        else:
            pass



from __future__ import annotations

from typing import Callable, List, Tuple, Union

import numpy as np

from robustnessgym.core.identifier import Identifier
from robustnessgym.core.operation import lookup
from robustnessgym.core.slice import SliceDataPanel as DataPanel
from robustnessgym.ops.spacy import SpacyOp
from robustnessgym.slicebuilders.subpopulations.score import ScoreSubpopulation


class NumTokensSubpopulation(ScoreSubpopulation):
    """Subpopulation based on token length."""

    def __init__(
        self,
        intervals: List[Tuple[Union[int, float, str], Union[int, float, str]]],
        reduction_fn: Callable = np.sum,
        **kwargs,
    ):
        super(NumTokensSubpopulation, self).__init__(
            intervals=intervals,
            identifiers=[
                Identifier(
                    _name=self.__class__.__name__,
                    gte=interval[0],
                    lte=interval[1],
                    reduction_fn=reduction_fn,
                )
                for interval in intervals
            ],
            **kwargs,
        )

        # Assign the reduction fn
        self.reduction_fn = reduction_fn

    def score(
        self,
        batch: DataPanel,
        columns: List[str],
        *args,
        **kwargs,
    ) -> np.ndarray:

        # Length of each example, for each column
        try:
            lengths = [
                [len(doc) for doc in lookup(batch, SpacyOp, [col])] for col in columns
            ]
        except AttributeError:
            lengths = [[len(text.split()) for text in batch[col]] for col in columns]

        # Reduction over column key axis
        return self.reduction_fn(np.array(lengths), axis=0)


class NumCharsSubpopulation(ScoreSubpopulation):
    """Subpopulation based on character length."""

    def __init__(
        self,
        intervals: List[Tuple[Union[int, float, str], Union[int, float, str]]],
        reduction_fn: Callable = np.sum,
        **kwargs,
    ):
        super(NumCharsSubpopulation, self).__init__(
            intervals=intervals,
            identifiers=[
                Identifier(
                    _name=self.__class__.__name__,
                    gte=interval[0],
                    lte=interval[1],
                    reduction_fn=reduction_fn,
                )
                for interval in intervals
            ],
            **kwargs,
        )

        # Assign the reduction fn
        self.reduction_fn = reduction_fn

    def score(
        self,
        batch: DataPanel,
        columns: List[str],
        *args,
        **kwargs,
    ) -> np.ndarray:

        # Length of each example, for each column
        lengths = [[len(text) for text in batch[col]] for col in columns]

        # Reduction over column key axis
        return self.reduction_fn(np.array(lengths), axis=0)



from __future__ import annotations

import subprocess
from typing import Collection, List

from meerkat import AbstractCell
from meerkat.tools.lazy_loader import LazyLoader

from robustnessgym.core.operation import Operation
from robustnessgym.core.slice import SliceDataPanel as DataPanel

textblob = LazyLoader(
    "textblob",
    error="TextBlob not available for import. Install using "
    "\npip install textblob\npython -m textblob.download_corpora",
)


class LazyTextBlobCell(AbstractCell):
    def __init__(self, text: str):
        super(LazyTextBlobCell, self).__init__()
        self.text = text

    def get(self, *args, **kwargs):
        return textblob.TextBlob(self.text)

    @property
    def data(self) -> object:
        return self.text

    @classmethod
    def _state_keys(cls) -> Collection:
        return {"text"}

    def __repr__(self):
        snippet = f"{self.text[:15]}..." if len(self.text) > 20 else self.text
        return f"{self.__class__.__name__}({snippet})"


class LazyTextBlobOp(Operation):
    def __init__(self):
        self._download()
        super(LazyTextBlobOp, self).__init__()

    @staticmethod
    def _download():
        subprocess.call(["python", "-m", "textblob.download_corpora"])

    def process_batch(
        self,
        dp: DataPanel,
        columns: List[str],
        **kwargs,
    ) -> tuple:
        return ([LazyTextBlobCell(text) for text in dp[columns[0]]],)



# flake8: noqa
from .activation import ActivationOp
from .allen import (
    AllenConstituencyParsingOp,
    AllenDependencyParsingOp,
    AllenPredictionOp,
    AllenSemanticRoleLabelingOp,
)
from .bootleg import BootlegAnnotatorOp
from .similarity import SentenceSimilarityMatrixOp
from .spacy import SpacyOp
from .stanza import StanzaOp
from .strip_text import StripTextOp
from .textblob import LazyTextBlobOp



"""Import common classes."""
# flake8: noqa
from meerkat.provenance import set_provenance

from robustnessgym.logging.utils import (
    initialize_logging,
    set_logging_level,
    set_logging_level_for_imports,
)

initialize_logging()
set_provenance()


from robustnessgym.core.devbench import DevBench
from robustnessgym.core.identifier import Id, Identifier
from robustnessgym.core.model import HuggingfaceModel, LudwigModel, Model
from robustnessgym.core.operation import Operation, lookup
from robustnessgym.core.slice import SliceDataPanel as DataPanel
from robustnessgym.core.testbench import TestBench

# from robustnessgym.ops.allen import (
#     AllenConstituencyParsingOp,
#     AllenDependencyParsingOp,
#     AllenPredictionOp,
#     AllenSemanticRoleLabelingOp,
# )
# from robustnessgym.ops.bootleg import BootlegAnnotatorOp
# from robustnessgym.ops.similarity import (
#     RougeMatrix,
#     RougeScore,
#     SentenceSimilarityMatrixOp,
# )
# from robustnessgym.ops.spacy import SpacyOp
# from robustnessgym.ops.stanza import StanzaOp
# from robustnessgym.ops.strip_text import StripTextOp
# from robustnessgym.ops.textblob import LazyTextBlobOp
from robustnessgym.slicebuilders.attacks.textattack import TextAttack
from robustnessgym.slicebuilders.slicebuilder import (  # SliceBuilderCollection,
    SliceBuilder,
)
from robustnessgym.slicebuilders.subpopulations.constituency_overlap import (
    ConstituencyOverlapSubpopulation,
    ConstituencySubtreeSubpopulation,
    FuzzyConstituencySubtreeSubpopulation,
)

# from robustnessgym.slicebuilders.subpopulations.entity_frequency
# import EntityFrequency
from robustnessgym.slicebuilders.subpopulations.hans import (
    HansAdjectives,
    HansAdjectivesCompEnt,
    HansAdjectivesCompNonEnt,
    HansAdverbs,
    HansAdvsEntailed,
    HansAdvsNonEntailed,
    HansAllPhrases,
    HansCalledObjects,
    HansConjs,
    HansConstAdv,
    HansConstQuotEntailed,
    HansEntComplementNouns,
    HansFoodWords,
    HansIntransitiveVerbs,
    HansLocationNounsA,
    HansLocationNounsB,
    HansNonEntComplementNouns,
    HansNonEntQuotVerbs,
    HansNPSVerbs,
    HansNPZVerbs,
    HansPassiveVerbs,
    HansPastParticiples,
    HansPluralNouns,
    HansPluralNPZVerbs,
    HansPrepositions,
    HansQuestionEmbeddingVerbs,
    HansQuestions,
    HansReadWroteObjects,
    HansRelations,
    HansSingularNouns,
    HansToldObjects,
    HansTransitiveVerbs,
    HansUnderstoodArgumentVerbs,
    HansWonObjects,
)
from robustnessgym.slicebuilders.subpopulations.length import NumTokensSubpopulation
from robustnessgym.slicebuilders.subpopulations.lexical_overlap import (
    LexicalOverlapSubpopulation,
)
from robustnessgym.slicebuilders.subpopulations.phrase import (
    AhoCorasick,
    HasAllPhrases,
    HasAnyPhrase,
    HasComparison,
    HasDefiniteArticle,
    HasIndefiniteArticle,
    HasNegation,
    HasPhrase,
    HasPosessivePreposition,
    HasQuantifier,
    HasTemporalPreposition,
)
from robustnessgym.slicebuilders.subpopulations.score import (
    BinarySubpopulation,
    IntervalSubpopulation,
    PercentileSubpopulation,
    ScoreSubpopulation,
)
from robustnessgym.slicebuilders.subpopulations.similarity import (
    Abstractiveness,
    Dispersion,
    Distillation,
    Ordering,
    Position,
    RougeMatrixScoreSubpopulation,
    RougeScoreSubpopulation,
)
from robustnessgym.slicebuilders.transformations.eda import EasyDataAugmentation
from robustnessgym.slicebuilders.transformations.fairseq import FairseqBacktranslation
from robustnessgym.slicebuilders.transformations.nlpaug import NlpAugTransformation
from robustnessgym.slicebuilders.transformations.similarity import (
    RougeMatrixSentenceTransformation,
)
from robustnessgym.tasks.task import (
    BinaryNaturalLanguageInference,
    BinarySentiment,
    ExtractiveQuestionAnswering,
    NaturalLanguageInference,
    QuestionAnswering,
    SentimentClassification,
    Summarization,
    Task,
    TernaryNaturalLanguageInference,
)

from .slicebuilders.attack import Attack



from functools import partial
from typing import List, Sequence, Tuple

import numpy as np

from robustnessgym.core.identifier import Identifier
from robustnessgym.core.slice import SliceDataPanel as DataPanel
from robustnessgym.core.tools import strings_as_json
from robustnessgym.ops.similarity import RougeMatrix
from robustnessgym.ops.spacy import SpacyOp
from robustnessgym.slicebuilders.transformation import Transformation


class RougeMatrixSentenceTransformation(Transformation):
    def __init__(self, metric: Sequence[str], threshold: float):
        super(RougeMatrixSentenceTransformation, self).__init__(
            num_transformed=1,
            identifiers=Identifier.range(n=1, _name=self.__class__.__name__),
        )

        self.metric = metric
        self.threshold = threshold

    def apply(
        self,
        batch: DataPanel,
        columns: List[str],
        skeleton_batches: List[DataPanel],
        slice_membership: np.ndarray,
        *args,
        **kwargs,
    ) -> Tuple[List[DataPanel], np.ndarray]:
        assert len(columns) == 2

        # Retrieve the relevant Rouge matrices
        matrices = RougeMatrix.retrieve(
            batch=batch,
            columns=columns,
            proc_fns=partial(RougeMatrix.select, metric=self.metric),
        )[strings_as_json(columns)]

        # Find max value along each row, remove rows that have max value below a
        # threshold
        rows_to_keep = [
            (m / (m.sum(axis=0) + 1e-5)).max(axis=1) >= self.threshold for m in matrices
        ]

        # Fetch sentences for the first column
        sentences = SpacyOp.retrieve(
            batch=batch,
            columns=[columns[0]],
            proc_fns=SpacyOp.sentences,
        )[columns[0]]

        # Delete sentences
        new_sentences = [
            " ".join(np.array(sent)[rows_to_keep[i]])
            for i, sent in enumerate(sentences)
        ]

        # Store the augmented text in the skeleton batches
        for i, augmented in enumerate(new_sentences):
            skeleton_batches[0][columns[0]][i] = augmented

        return skeleton_batches, slice_membership



from typing import List

import numpy as np

from robustnessgym.core.operation import lookup
from robustnessgym.core.slice import SliceDataPanel as DataPanel
from robustnessgym.ops.spacy import SpacyOp
from robustnessgym.slicebuilders.subpopulations.score import ScoreSubpopulation


class LexicalOverlapSubpopulation(ScoreSubpopulation):
    def score(
        self,
        batch: DataPanel,
        columns: List[str],
        *args,
        **kwargs,
    ) -> np.ndarray:
        # Require that the number of keys is exactly 2
        assert len(columns) == 2, "Must specify exactly 2 keys."

        # Lookup the tokens after lower-casing and placing into a set
        try:
            tokens = {
                col: [
                    set([str(tok).lower() for tok in doc])
                    for doc in lookup(batch, SpacyOp, [col])
                ]
                for col in columns
            }
        except AttributeError:
            tokens = {
                col: [
                    set([str(tok).lower() for tok in text.split()])
                    for text in batch[col]
                ]
                for col in columns
            }

        # Compute the intersection over union score
        return np.array(
            [
                len(tokens_0.intersection(tokens_1))
                / float(len(tokens_0.union(tokens_1)))
                for tokens_0, tokens_1 in zip(tokens[columns[0]], tokens[columns[1]])
            ]
        )



from typing import List

from meerkat.tools.lazy_loader import LazyLoader

from robustnessgym.core.operation import Operation
from robustnessgym.core.slice import SliceDataPanel as DataPanel
from robustnessgym.mixins.device import DeviceMixin

predictors = LazyLoader("allennlp.predictors")


class AllenPredictionOp(DeviceMixin, Operation):
    def __init__(
        self,
        path: str,
        device: str,
    ):
        super(AllenPredictionOp, self).__init__(device=device)

        self._predictor = predictors.Predictor.from_path(
            archive_path=path, cuda_device=self.cuda_device
        )

    @property
    def predictor(self):
        return self._predictor

    def process_batch(
        self,
        dp: DataPanel,
        columns: List[str],
        **kwargs,
    ) -> tuple:
        return (
            self.predictor.predict_batch_json(
                [{"sentence": text} for text in dp[columns[0]]]
            ),
        )


class AllenConstituencyParsingOp(AllenPredictionOp):
    def __init__(self, device: str = None):
        super(AllenConstituencyParsingOp, self).__init__(
            path="https://storage.googleapis.com/allennlp-public-models/elmo"
            "-constituency-parser-2020.02.10.tar.gz",
            device=device,
        )

    def process_batch(
        self,
        dp: DataPanel,
        columns: List[str],
        **kwargs,
    ) -> tuple:
        return (
            [
                p["trees"]
                for p in self.predictor.predict_batch_json(
                    [{"sentence": text} for text in dp[columns[0]]]
                )
            ],
        )


class AllenDependencyParsingOp(AllenPredictionOp):
    def __init__(self, device: str = None):
        super(AllenDependencyParsingOp, self).__init__(
            path="https://storage.googleapis.com/allennlp-public-models/"
            "biaffine-dependency-parser-ptb-2020.04.06.tar.gz",
            device=device,
        )


class AllenSemanticRoleLabelingOp(AllenPredictionOp):
    def __init__(self, device: str = None):
        super(AllenSemanticRoleLabelingOp, self).__init__(
            path="https://storage.googleapis.com/allennlp-public-models/bert-base-srl"
            "-2020.03.24.tar.gz",
            device=device,
        )



# flake8: noqa
import glob
import inspect
from pathlib import Path

import numpy as np
import pyarrow
import streamlit as st
from meerkat.columns.prediction_column import ClassificationOutputColumn

import robustnessgym as rg
from robustnessgym import DataPanel


# @st.cache(hash_funcs={pyarrow.lib.Buffer: lambda x: 0})
def load_boolq():
    return DataPanel.load_huggingface(
        "boolq",
        split="validation",
    )


# @st.cache#(hash_funcs={pyarrow.lib.Buffer: lambda x: 0})
def load_boolq_small():
    return DataPanel.load_huggingface("boolq", split="validation[:2]")


def load_datapanel(dataset="boolq", size="small"):
    if dataset == "boolq":
        if size == "small":
            return load_boolq_small()
        else:
            return load_boolq()


def format_code(code=None, output=None, outputs=None, columns=(0.45, 0.1, 0.45)):
    with st.beta_container():
        code_col, _, stdout_col = st.beta_columns(columns)
        with code_col:
            st.code(code)
        with stdout_col:
            if output is not None:
                st.write(output)
            else:
                for output in outputs:
                    st.write(output)
    st.write("<br>", unsafe_allow_html=True)


def datapanel_page(dataset="boolq", size="small"):
    st.write("# DataPanels: Get your data into Robustness Gym")
    st.write(
        """
Load data into Robustness Gym using the `DataPanel`. For detailed info about 
the `DataPanel`, check out [Mosaic](https://github.com/robustness-gym/meerkat).
        """
    )

    dp = display_dp(dataset, size)

    st.write(
        """
The `DataPanel` organizes data into columns, where columns have heterogenous types.
For example, in the `DataPanel` we loaded, some columns are of type `ListColumn`,
while one is of type `NumpyArrayColumn`. 
        """
    )

    datapanel_sheet(dp)


def datapanel_sheet(dp):
    with st.beta_container():
        code_col, _, stdout_col = st.beta_columns([0.30, 0.05, 0.65])
        with code_col:
            st.write("## Code Snippet")
        with stdout_col:
            st.write("## Output")

    format_code("dp", dp.streamlit(), columns=[0.30, 0.05, 0.65])
    format_code(
        """
dp['passage']
type(dp['passage'])
        """,
        outputs=(dp["passage"].streamlit(), str(type(dp["passage"]))),
        columns=[0.30, 0.05, 0.65],
    )
    format_code(
        """
dp['answer']
type(dp['answer'])
        """,
        outputs=(dp["answer"].streamlit(), str(type(dp["answer"]))),
        columns=[0.30, 0.05, 0.65],
    )
    format_code(
        """
# NumpyArrayColumn works like np.ndarray!
dp['answer'].mean()
dp['answer'].reshape(-1, 1, 1)
        """
        "",
        outputs=(dp["answer"].mean(), dp["answer"].reshape(-1, 1, 1).streamlit()),
        columns=[0.30, 0.05, 0.65],
    )
    format_code("dp[0]", dp[0], columns=[0.30, 0.05, 0.65])
    format_code("dp[:1]", dp[:1].streamlit(), columns=[0.30, 0.05, 0.65])
    format_code("dp.identifier", dp.identifier, columns=[0.30, 0.05, 0.65])
    format_code("dp.columns", dp.columns, columns=[0.30, 0.05, 0.65])
    format_code(
        "dp.map(lambda x: len(x['passage']))",
        dp.map(lambda x: len(x["passage"])).streamlit(),
        columns=[0.30, 0.05, 0.65],
    )
    format_code(
        """
dp.map(lambda x: {
    'passage_len': len(x['passage']),  
    'question_len': len(x['question']),
})
        """,
        dp.map(
            lambda x: {
                "passage_len": len(x["passage"]),
                "question_len": len(x["question"]),
            }
        ).streamlit(),
        columns=[0.30, 0.05, 0.65],
    )
    format_code(
        "dp.filter(lambda x: len(x['passage']) > 600)",
        dp.filter(lambda x: len(x["passage"]) > 600).streamlit(),
        columns=[0.30, 0.05, 0.65],
    )
    format_code(
        "dp.update(lambda x: {'passage_len': len(x['passage'])})",
        dp.update(lambda x: {"passage_len": len(x["passage"])}).streamlit(),
        columns=[0.30, 0.05, 0.65],
    )
    format_code(
        """
dp.update(lambda x: {
    'passage_len': len(x['passage']),  
    'question_len': len(x['question']),
})
        """,
        dp.update(
            lambda x: {
                "passage_len": len(x["passage"]),
                "question_len": len(x["question"]),
            }
        ).streamlit(),
        columns=[0.30, 0.05, 0.65],
    )
    format_code("dp.to_pandas()", dp.to_pandas(), columns=[0.30, 0.05, 0.65])

    st.write("------")


def display_dp(dataset="boolq", size="small"):
    st.subheader("Load Data")
    code_col, stdout_col = st.beta_columns(2)
    with code_col:
        if dataset == "boolq":
            if size == "small":
                dp = load_boolq_small()
                st.code(
                    """
# use the DataPanel to get your data into RG
from robustnessgym import DataPanel  
# any Huggingface dataset
dp = DataPanel.load_huggingface('boolq', split='validation[:2]')  
                    """
                )
            else:
                dp = load_boolq()
                st.code(
                    """
# use the DataPanel to get your data into RG
from robustnessgym import DataPanel
# any Huggingface dataset
dp = DataPanel.load_huggingface('boolq', split='validation')
                    """
                )

    return dp


def operation_page():
    st.write("# Operations: Run common processing workflows in Robustness Gym")
    st.write(
        """
An `Operation` in Robustness Gym is used to write and run common workflows 
e.g. tokenizing text, computing embeddings, tagging text, etc. 
Run an Operation on a `DataPanel` to
add new columns to the `DataPanel` that can be used for analysis. 

The main thing to remember: an `Operation` adds new columns to your `DataPanel`.
        """
    )

    dp = display_dp()

    with st.beta_expander("Run the spaCy pipeline with `SpacyOp`"):
        spacy_example(dp)

    with st.beta_expander("Run the Stanza pipeline with `StanzaOp`"):
        stanza_example(dp)

    with st.beta_expander("Lazily run the TextBlob pipeline with `LazyTextBlobOp`"):
        lazy_textblob_example(dp)

    with st.beta_expander("Write a simple `Operation` to capitalize text"):
        custom_operation_example_1(dp)

    with st.beta_expander(
        "Write an `Operation` that adds "
        "multiple columns to capitalize and upper-case text"
    ):
        custom_operation_example_2(dp)


def spacy_example(dp):
    columns = [0.45, 0.05, 0.50]

    with st.beta_container():
        st.header("Run Operation: `SpacyOp`")
        st.write(
            """
spaCy is a popular text processing library that provides tokenization, tagging 
and other capabilities. 
            """
        )

        from robustnessgym import lookup
        from robustnessgym.ops import SpacyOp

        # Run the Spacy pipeline on the 'question' column of the dataset
        spacy = SpacyOp()
        dp = spacy(dp=dp, columns=["passage"])
        # adds a new column that is auto-named
        # "SpacyOp(lang=en_core_web_sm, neuralcoref=False, columns=['passage'])"

        format_code(
            """
from robustnessgym import lookup
from robustnessgym.ops import SpacyOp

# Run the Spacy pipeline on the 'question' column of the dataset
spacy = SpacyOp()
dp = spacy(dp=dp, columns=['passage'])
# adds a new column that is auto-named
# "SpacyOp(lang=en_core_web_sm, neuralcoref=False, columns=['passage'])"
            """,
            dp.streamlit(),
            columns=columns,
        )

    st.write("------")

    with st.beta_container():
        spacy_column = lookup(dp, spacy, ["passage"])
        format_code(
            """
lookup(dp, spacy, ['passage'])
            """,
            spacy_column._repr_pandas_(),
            columns=columns,
        )

    with st.beta_container():
        st.subheader("Columns contain Cells")
        cell = spacy_column[1]
        format_code(
            """
cell = spacy_column[1]
cell
            """,
            cell,
            columns=columns,
        )

        format_code(
            """
list(cell)
            """,
            list(cell),
            columns=columns,
        )
        format_code(
            """
cell.ents
            """,
            cell.ents,
            columns=columns,
        )


def stanza_example(dp):
    columns = [0.45, 0.05, 0.50]
    st.header("Run Stanza Workflow")
    from robustnessgym import lookup
    from robustnessgym.ops import StanzaOp

    # Run the Stanza pipeline on the 'question' column of the dataset
    stanza = StanzaOp()
    dp = stanza(dp=dp, columns=["question"])
    # adds a new column that is auto-named "StanzaOp(columns=['question'])"

    # Grab the Stanza column from the DataPanel using the lookup
    stanza_column = lookup(dp, stanza, ["question"])
    format_code(
        """
from robustnessgym import lookup
from robustnessgym.ops import StanzaOp

# Run the Stanza pipeline on the 'question' column of the dataset
stanza = StanzaOp()
dp = stanza(dp=dp, columns=['question'])
# adds a new column that is auto-named "StanzaOp(columns=['question'])"

# Grab the Stanza column from the DataPanel using the lookup
stanza_column = lookup(dp, stanza, ['question'])
        """,
        stanza_column._repr_pandas_(),
        columns=columns,
    )
    st.subheader("Columns contain Cells")
    format_code(
        """ 
cell = stanza_column[0]
cell
        """,
        stanza_column[0],
        columns=columns,
    )
    st.subheader("Cells can be treated like stanza objects")
    format_code(
        """ 
cell = stanza_column[0]
cell.text
        """,
        stanza_column[0].text,
        columns=columns,
    )
    format_code(
        """ 
cell = stanza_column[0]
cell.entities
        """,
        stanza_column[0].entities,
        columns=columns,
    )


def lazy_textblob_example(dp):
    columns = [0.45, 0.05, 0.50]
    st.header("Run TextBlob Workflow Lazily")

    def fn():
        nonlocal dp
        from robustnessgym import lookup
        from robustnessgym.ops import LazyTextBlobOp

        # Run the TextBlob pipeline on the 'passage' column of the dataset
        textblob = LazyTextBlobOp()
        dp = textblob(dp=dp, columns=["question"])
        # adds a new column that is auto-named "LazyTextBlobOp(columns=['question'])"

        # Grab the TextBlob column from the DataPanel using the lookup
        textblob_column = lookup(dp, textblob, ["question"])
        return textblob_column

    fn_source = inspect.getsource(fn)
    fn_source = "\n".join([line[4:] for line in fn_source.split("\n")[2:]])
    textblob_column = fn()
    format_code(fn_source, textblob_column._repr_pandas_(), columns=columns)

    st.subheader("Columns contain cells")
    cell = textblob_column[1]
    format_code(
        """
cell = textblob_column[1]
cell
        """,
        textblob_column[1],
        columns=columns,
    )
    format_code(
        """
cell.sentences
        """,
        cell.sentences,
        columns=columns,
    )


def custom_operation_example_1(dp):
    columns = [0.45, 0.05, 0.50]
    st.header("Run Custom Operation")

    def fn():
        from robustnessgym import Id, Operation

        # A function that capitalizes text
        def capitalize(batch, columns):
            return [text.capitalize() for text in batch[columns[0]]]

        # Wrap in an Operation: `process_batch_fn` accepts functions that have
        # exactly 2 arguments: batch and columns, and returns a tuple of outputs
        op = Operation(
            identifier=Id("CapitalizeOp"),
            process_batch_fn=capitalize,
        )
        return op

    fn_source = inspect.getsource(fn)
    fn_source = "\n".join([line[4:] for line in fn_source.split("\n")[1:]])
    op = fn()
    format_code(fn_source, op, columns=columns)

    def fn():
        nonlocal dp
        from robustnessgym import lookup

        dp = op(dp=dp, columns=["question"])

        # Look it up when you need it
        capitalized_text = lookup(dp, op, ["question"])
        return capitalized_text

    fn_source = inspect.getsource(fn)
    fn_source = "\n".join([line[4:] for line in fn_source.split("\n")[1:]])
    capitalized_text = fn()
    format_code(fn_source, capitalized_text._repr_pandas_(), columns=columns)

    format_code("capitalized_text[0]", capitalized_text[0], columns=columns)


def custom_operation_example_2(dp):
    columns = [0.45, 0.05, 0.50]
    st.header("Run Custom Operation")

    def fn():
        from robustnessgym import Id, Operation

        # A function that capitalizes and upper-cases text: this will
        # be used to add two columns to the DataPanel
        def capitalize_and_upper(batch, columns):
            return [text.capitalize() for text in batch[columns[0]]], [
                text.upper() for text in batch[columns[0]]
            ]

        # Wrap in an Operation: `process_batch_fn` accepts functions that have
        # exactly 2 arguments: batch and columns, and returns a tuple of outputs
        op = Operation(
            identifier=Id("ProcessingOp"),
            output_names=["capitalize", "upper"],
            process_batch_fn=capitalize_and_upper,
        )
        return op

    fn_source = inspect.getsource(fn)
    fn_source = "\n".join([line[4:] for line in fn_source.split("\n")[1:]])
    op = fn()
    format_code(fn_source, op, columns=columns)

    def fn():
        nonlocal dp
        # Apply to a DataPanel
        dp = op(dp=dp, columns=["question"])
        return dp

    fn_source = inspect.getsource(fn)
    fn_source = "\n".join([line[4:] for line in fn_source.split("\n")[2:]])
    new_dp = fn()
    format_code(fn_source, new_dp._repr_pandas_(), columns=columns)

    def fn():
        nonlocal dp, op
        from robustnessgym import lookup

        capitalized_text = lookup(dp, op, ["question"], "capitalize")
        return capitalized_text

    fn_source = inspect.getsource(fn)
    fn_source = "\n".join([line[4:] for line in fn_source.split("\n")[2:]])
    capitalized_text = fn()
    format_code(fn_source, capitalized_text._repr_pandas_(), columns=columns)

    def fn():
        nonlocal dp, op
        from robustnessgym import lookup

        upper_text = lookup(dp, op, ["question"], "upper")
        return upper_text

    fn_source = inspect.getsource(fn)
    fn_source = "\n".join([line[4:] for line in fn_source.split("\n")[2:]])
    upper_text = fn()
    format_code(fn_source, upper_text._repr_pandas_(), columns=columns)


def subpopulation_example_1():

    dp = display_dp(size="all")

    col1, col2 = st.beta_columns(2)

    with col2:
        with st.echo():
            from robustnessgym import DataPanel, ScoreSubpopulation, lookup
            from robustnessgym.ops import SpacyOp

            def length(batch: DataPanel, columns: list):
                try:
                    # Take advantage of previously stored Spacy information
                    return [len(doc) for doc in lookup(batch, SpacyOp, columns)]
                except AttributeError:
                    # If unavailable, fall back to splitting text
                    return [len(text.split()) for text in batch[columns[0]]]

            # Create a subpopulation that buckets examples based on length
            length_sp = ScoreSubpopulation(
                intervals=[(0, 10), (10, 20)],
                score_fn=length,
            )

            slices, membership = length_sp(dp=dp, columns=["passage"])
            # `slices` is a list of 2 DataPanels
            # `membership` is an np.ndarray of shape (n x 2)
            st.markdown("`slices`")
            st.write(slices)
            st.markdown("`membership`")
            st.write(membership)

        with st.echo():
            # Create a subpopulation that buckets examples based on length
            length_sp = ScoreSubpopulation(
                # mixture of percentile intervals and raw intervals
                intervals=[("0%", "10%"), ("10%", "20%"), (10, 20)],
                score_fn=length,
            )

            slices, membership = length_sp(dp=dp, columns=["passage"])
            st.write("`slices` sizes:", [len(sl) for sl in slices])
            st.write("`membership` shape:", membership.shape)

        st.write("### Updated intervals")
        with st.echo():
            # The percentile intervals are updated to
            # raw values after running the subpopulation
            st.write(length_sp.intervals)

        st.write("### Examples and average passage length for each slice")
        with st.echo():
            for sl in slices:
                # The slice identifier
                st.write(str(sl.identifier))
                # Look at the first 3 examples from the slice
                st.write(sl.head(3)._repr_pandas_())
                # Calculate the average length of the passages in the slice
                st.write(
                    sl.map(lambda x: length(x, ["passage"]), is_batched_fn=True).mean()
                )


EXAMPLE_DATASETS = [
    (
        "snli",
        [
            "validation",  # "test"
        ],
    ),
    ("hans", ["validation"]),
    # ("anli",
    #  ["dev_r1", "test_r1",
    #   "dev_r2", "test_r2",
    #   "dev_r3", "test_r3"
    #   ]),
    (
        "glue/mnli",
        [
            "validation_matched",  # "validation_mismatched", "test_matched", "test_mismatched"
        ],
    ),
]
EXAMPLE_DATASETS_TO_SPLITS = {x[0]: x[1] for x in EXAMPLE_DATASETS}

EXAMPLE_MODELS = [
    "ynie/roberta-large-snli_mnli_fever_anli_R1_R2_R3-nli",
    "ynie/albert-xxlarge-v2-snli_mnli_fever_anli_R1_R2_R3-nli",
    "ynie/bart-large-snli_mnli_fever_anli_R1_R2_R3-nli",
    "ynie/xlnet-large-cased-snli_mnli_fever_anli_R1_R2_R3-nli",
    "textattack/bert-base-uncased-snli",
    "textattack/distilbert-base-cased-snli",
    "textattack/bert-base-uncased-MNLI",
    "textattack/albert-base-v2-snli",
    "facebook/bart-large-mnli",
    "roberta-large-mnli",
    "microsoft/deberta-large-mnli",
    "microsoft/deberta-v2-xxlarge-mnli",
    "typeform/mobilebert-uncased-mnli",
    "huggingface/distilbert-base-uncased-finetuned-mnli",
    "prajjwal1/albert-base-v1-mnli",
    "prajjwal1/bert-tiny-mnli",
    "cross-encoder/nli-deberta-base",
    "squeezebert/squeezebert-mnli",
    "squeezebert/squeezebert-mnli-headless",
]


def select_dataset():
    with st.beta_container():
        col1, col2 = st.beta_columns(2)
        with col1:
            dataset = st.radio(
                "Dataset",
                [t[0] for t in EXAMPLE_DATASETS],
            )
        with col2:
            split = st.radio(
                "Split",
                EXAMPLE_DATASETS_TO_SPLITS[dataset],
            )

    return load_dataset(dataset, split)


# @st.cache(allow_output_mutation=True)
def load_dataset(dataset, split):
    dp = DataPanel.load_huggingface(*dataset.split("/"), split=split)
    return dataset.replace("/", "_"), split, dp


# @st.cache(allow_output_mutation=True)
def load_predictions(dataset, split, dp, dir="nli-preds"):
    filepaths = [Path(path) for path in glob.glob(f"{dir}/*")]

    filepaths = [
        path for path in filepaths if path.name.startswith(f"{dataset}-{split}")
    ]

    model_info = {
        path.name.replace(f"{dataset}-{split}-", ""): ClassificationOutputColumn.read(
            str(path)
        )
        for path in filepaths
    }

    # for model, preds in model_info.items():
    #     dp.add_column(model, preds, overwrite=True)

    for model, preds in model_info.items():
        if (
            model.startswith("huggingface")
            or model.startswith("textattack-bert-base")
            or model.startswith("prajjwal")
            or model.startswith("squeezebert-squeezebert")
            or model.startswith("facebook")
        ):
            preds = preds.preds().map(lambda x: {0: 2, 1: 0, 2: 1}[x.item()])
        else:
            preds = preds.preds().map(lambda x: {0: 0, 1: 1, 2: 2}[x.item()])
        dp.add_column(model, preds, overwrite=True)

    return model_info, filepaths, dp


def run_demo():
    # Pick a dataset
    dataset, split, dp = select_dataset()

    st.header(dataset)
    st.subheader(split)
    # st.write(dp.streamlit())

    # Load predictions
    model_info, filepaths, dp = load_predictions(dataset, split, dp)
    # st.write(dp.streamlit())

    # # Pick models
    # models = st.multiselect("Model(s)",
    #                         [model for model in EXAMPLE_MODELS
    #                          if model.replace("/", "-") in model_info])

    with st.beta_container():
        col1, col2 = st.beta_columns(2)
        with col1:
            model_1 = st.selectbox(
                "Model 1",
                [
                    model
                    for model in EXAMPLE_MODELS
                    if model.replace("/", "-") in model_info
                ],
            )
        with col2:
            model_2 = st.selectbox(
                "Model 2",
                [
                    model
                    for model in EXAMPLE_MODELS
                    if model.replace("/", "-") in model_info
                ],
                index=1,
            )

    # if not models:
    #     st.stop()
    from functools import partial

    def accuracy(x, model):
        return (x[model] == x["label"]).mean()

    db = rg.DevBench()
    db.add_aggregators(
        {
            model: {
                "accuracy": partial(accuracy, model=model),
            }
            for model in model_info
        }
    )

    # st.multiselect(
    #     "Slice(s)",
    #     [rg.LexicalOverlapSubpopulation([('0%', '10%'), ('90%', '100%')])]
    # )

    intervals = [("0%", "10%"), ("20%", "80%"), ("90%", "100%")]
    sps = [rg.HasNegation(), rg.NumTokensSubpopulation(intervals=intervals),] + [
        rg.HansAdverbs(),
        rg.HansConjs(),
        rg.HansAdjectives(),
        rg.HansAdjectivesCompNonEnt(),
        rg.HansAdjectivesCompEnt(),
        rg.HansCalledObjects(),
        rg.HansAdvsEntailed(),
        rg.HansAdvsNonEntailed(),
        rg.HasTemporalPreposition(),
        rg.HasComparison(),
        rg.HasQuantifier(),
        rg.HasPosessivePreposition(),
        rg.HansSingularNouns(),
        rg.HasDefiniteArticle(),
        rg.HasIndefiniteArticle(),
    ]
    # dual_sps = [
    #     rg.LexicalOverlapSubpopulation(intervals),
    # ]

    slices = st.multiselect("Slice(s)", sps, format_func=lambda x: x.__class__.__name__)

    generate_report = True  # st.button("Generate Report")

    if generate_report:
        db.add_slices(dp)

        for sl in slices:
            db(sl, dp, ["premise"])
            db(sl, dp, ["hypothesis"])

        db(
            rg.LexicalOverlapSubpopulation(intervals),
            dp,
            ["premise", "hypothesis"],
        )
        # st.write(db.create_report([model.replace("/", "-") for model in models]).figure())
        # for model in models:

        report = db.create_report([model_1.replace("/", "-")])
        report.update_config(**dict(color_scheme=["#3499ec", "#ec34c1", "#9cec34"]))
        st.write(report.figure())
        st.write(db.create_report([model_2.replace("/", "-")]).figure())

    # for model in models:
    #     st.write(model_info[model.replace("/", "-")].preds().to_pandas())


def transformation_example_1():
    col1, col2 = st.beta_columns(2)
    dp = display_dp(col1)


if __name__ == "__main__":
    st.set_page_config(layout="wide")

    # st.title("RG Ops Demo")

    st.sidebar.title(":rocket: Robustness Gym\n " "## Cheat Sheet")

    st.sidebar.write("#### Installation")
    st.sidebar.code(
        """
$ pip install robustnessgym
# see setup.py for a list of optional packages
$ pip install robustnessgym[text]
        """,
        language="bash",
    )

    st.sidebar.write("#### Import convention")
    st.sidebar.code("import robustnessgym as rg")

    st.sidebar.write("### Selection")
    section = st.sidebar.radio(
        "Section",
        ["DataPanel", "Operation", "Subpopulation", "Demo"],
    )

    if section == "DataPanel":
        datapanel_page(size="all")

    elif section == "Operation":
        operation_page()

    elif section == "Subpopulation":
        # selection = st.sidebar.radio("Subpopulation Example", ["Subpopulation-1"])
        # if selection == "Subpopulation-1":
        subpopulation_example_1()

    elif section == "Transformation":
        selection = st.sidebar.selectbox("Transformation Example", ["Transformation-1"])
        if selection == "Transformation-1":
            transformation_example_1()

    elif section == "Demo":
        run_demo()



"""Example of running predictions on NLI tasks.

Example:

    # Run squeezebert and bart large (facebook) from HuggingFace on boolq
    $ python prediction.py --dataset boolq \
        --model squeezebert/squeezebert-mnli facebook/bart-large-mnli \
        --output-dir ./
"""
import argparse
import itertools
import os

import pandas as pd
from meerkat import DataPanel
from meerkat.columns.prediction_column import ClassificationOutputColumn

from robustnessgym.core.model import Model
from robustnessgym.tasks.task import BinaryNaturalLanguageInference

# EXAMPLE_DATASETS = [
#     ("boolq", "validation"),
#     ("snli", "validation"),
#     (("glue", "mnli"), "validation_matched"),
# ]

EXAMPLE_DATASETS = [
    ("hans", ["validation"]),
    ("snli", ["validation", "test"]),
    ("anli", ["dev_r1", "test_r1", "dev_r2", "test_r2", "dev_r3", "test_r3"]),
    (
        ("glue", "mnli"),
        [
            "validation_matched",
            "validation_mismatched",
            "test_matched",
            "test_mismatched",
        ],
    ),
]
EXAMPLE_DATASETS_TO_SPLITS = {x[0]: x[1] for x in EXAMPLE_DATASETS}

EXAMPLE_MODELS = [
    "ynie/roberta-large-snli_mnli_fever_anli_R1_R2_R3-nli",
    "ynie/albert-xxlarge-v2-snli_mnli_fever_anli_R1_R2_R3-nli",
    "ynie/bart-large-snli_mnli_fever_anli_R1_R2_R3-nli",
    "ynie/xlnet-large-cased-snli_mnli_fever_anli_R1_R2_R3-nli",
    "textattack/bert-base-uncased-snli",
    "textattack/distilbert-base-cased-snli",
    "textattack/bert-base-uncased-MNLI",
    "textattack/albert-base-v2-snli",
    "facebook/bart-large-mnli",
    "roberta-large-mnli",
    "microsoft/deberta-large-mnli",
    "microsoft/deberta-v2-xxlarge-mnli",
    "typeform/mobilebert-uncased-mnli",
    "huggingface/distilbert-base-uncased-finetuned-mnli",
    "prajjwal1/albert-base-v1-mnli",
    "prajjwal1/bert-tiny-mnli",
    "cross-encoder/nli-deberta-base",
    "squeezebert/squeezebert-mnli",
    "squeezebert/squeezebert-mnli-headless",
]


def run_nli(
    dataset_info,
    model_name,
    debug: bool = False,
    output_dir: str = None,
):
    """Compute and save logits for natural language inference tasks.

    Args:
        dataset_info (Tuple[Union[str, Tuple[str]]]): Dataset name and split.
        model_name (str): The HuggingFace model to use.
        debug (bool, optional): If ``True``, runs in debug mode.
            This runs inference on the first 10 examples in the dataset.

    Returns:
        ClassificationOutputColumn: The output column with DataPanel.
    """

    def fn(dp):
        """Computes logits over the datapanel."""
        if "passage" in dp:
            data = {"col1": dp["passage"].data, "col2": dp["question"].data}
        else:
            data = {"col1": dp["premise"].data, "col2": dp["hypothesis"].data}

        output = model.predict_batch(data, ["col1", "col2"])  # noqa: F821
        return output["logits"].cpu()

    # Load the dataset as a DataPanel.
    dataset_name, split = dataset_info
    if isinstance(dataset_name, str):
        dataset_name = (dataset_name,)
    dp = DataPanel.load_huggingface(*dataset_name, split=split)
    if debug:
        dp = dp[:10]

    # Load the model.
    model = Model.huggingface(
        model_name, task=BinaryNaturalLanguageInference(), is_classifier=True
    )

    # Compute predictions on the data.
    out = dp.map(
        fn,
        batch_size=5,
        is_batched_fn=True,
        output_type=ClassificationOutputColumn,
        pbar=True,
    )
    out = ClassificationOutputColumn(logits=out)

    del model

    model_name = model_name.replace("/", "-")
    if output_dir:
        dirname = (
            f"{output_dir}/nli-preds/" f"{'_'.join(dataset_name)}-{split}-{model_name}"
        )
        out.write(dirname)

    # TO LOAD:
    # ClassificationOutputColumn.read(dirname)

    return out


def parse_args():
    parser = argparse.ArgumentParser(
        "Run NLI tasks in RobustnessGym with built-in HuggingFace models."
    )
    parser.add_argument(
        "--dataset",
        nargs="+",
        type=str,
        choices=[
            "_".join(x[0]) if isinstance(x[0], tuple) else x[0]
            for x in EXAMPLE_DATASETS
        ],
        help="Dataset(s) to use",
        required=True,
    )
    parser.add_argument(
        "--model",
        nargs="+",
        type=str,
        choices=EXAMPLE_MODELS,
        help="Model(s) to use",
    )
    parser.add_argument(
        "--output-dir",
        type=str,
        help="Output directory to save predictions for easy loading",
        default="./",
    )
    parser.add_argument("--debug", action="store_true", help="Run in debug mode")
    return parser.parse_args()


def main():
    args = parse_args()
    datasets = [tuple(x.split("_")) for x in args.dataset]
    datasets = [x if len(x) > 1 else x[0] for x in datasets]
    all_models = args.model
    output_dir = args.output_dir
    debug = args.debug

    if all_models is None:
        all_models = EXAMPLE_MODELS

    configs = list(itertools.product(datasets, all_models))
    outs = []
    for idx, (ds, model) in enumerate(configs):
        for split in EXAMPLE_DATASETS_TO_SPLITS[ds]:
            print(
                f"***** ({idx}/{len(configs)}: "
                f"Dataset: {ds} - Split: {split} - Model: {model}) *****"
            )
            ds = (ds, split)
            try:
                run_nli(ds, model, debug=debug, output_dir=output_dir)
                outs.append(
                    {
                        "dataset": str(ds),
                        "model": model,
                        "status": "Success",
                        "Reason": None,
                    }
                )
            except Exception as e:
                print(e)
                outs.append(
                    {
                        "dataset": str(ds),
                        "model": model,
                        "status": "Failed",
                        "Reason": str(e),
                    }
                )

    if output_dir:
        run_status = pd.DataFrame(outs)
        run_status.to_csv(os.path.join(output_dir, "run_status.csv"))


if __name__ == "__main__":
    main()



from __future__ import annotations

from typing import Dict, List

from meerkat import AbstractCell
from meerkat.tools.lazy_loader import LazyLoader

from robustnessgym.core.operation import Operation
from robustnessgym.core.slice import SliceDataPanel as DataPanel

stanza = LazyLoader("stanza", error="Please `pip install stanza`.")


class StanzaCell(AbstractCell):
    """Cell that stores a Stanza Document."""

    def __init__(self, doc: stanza.Document, *args, **kwargs):
        super(StanzaCell, self).__init__(*args, **kwargs)
        self.doc = doc

    def get(self, *args, **kwargs):
        return self.doc

    @property
    def data(self) -> stanza.Document:
        return self.doc

    @classmethod
    def from_text(cls, text: str, pipeline: stanza.Pipeline) -> StanzaCell:
        return cls(pipeline(text))

    def get_state(self) -> Dict:
        return {"doc": self.doc.to_serialized()}

    @classmethod
    def from_state(cls, state, **kwargs) -> StanzaCell:
        return cls(stanza.Document.from_serialized(state["doc"]))

    def __repr__(self):
        snippet = (
            f"{self.doc.text[:15]}..." if len(self.doc.text) > 20 else self.doc.text
        )
        return f"{self.__class__.__name__}({snippet})"

    def _repr_html_(self):
        return self.__repr__()

    def __getattr__(self, item):
        try:
            if item in {
                "lemma",
                "text",
                "upos",
                "xpos",
                "feats",
                "head",
                "deprel",
                "feats",
                "misc",
                "id",
            }:
                return self.doc.get(item)
        except AttributeError:
            return super().__getattr__(item)


class StanzaOp(Operation):
    """Operation that runs the Stanza pipeline.

    Stanza: https://stanfordnlp.github.io/stanza/
    """

    def __init__(self):
        super(StanzaOp, self).__init__()

        self._download()
        self.nlp = stanza.Pipeline()

    @staticmethod
    def _download():
        stanza.download()

    def process_batch(
        self,
        dp: DataPanel,
        columns: List[str],
        **kwargs,
    ) -> tuple:
        """Process text examples by running them through the Stanza pipeline.

        Args:
            dp (DataPanel): DataPanel
            columns (list): list of columns
            **kwargs: optional keyword arguments

        Returns:
            Tuple with single output, a list of StanzaCell objects.
        """

        return tuple(
            [
                [StanzaCell.from_text(text, self.nlp) for text in dp[col]]
                for col in columns
            ]
        )


# def input_columns(self):
#     return ['image', 'bb']

# op(dp, image_column=['image'], [{'image': '])

# op = StanzaOp()
# op(dp, 'passage') --> 1 output column
# op(dp, ['passage', 'question']) --> 2 output columns



from functools import reduce
from typing import List

import torch.nn as nn

from robustnessgym.core.operation import Operation
from robustnessgym.core.slice import SliceDataPanel as DataPanel


def _nested_getattr(obj, attr, *args):
    """Get a nested property from an object.

    Example:
    ```
        model = ...
        weights = _nested_getattr(model, "layer4.weights")
    ```
    """
    return reduce(lambda o, a: getattr(o, a, *args), [obj] + attr.split("."))


class ActivationExtractor:
    """Class for extracting activations of a targeted intermediate layer."""

    def __init__(self):
        self.activation = None

    def add_hook(self, module, input, output):
        self.activation = output


# TODO: test on an NLP model
class ActivationOp(Operation):
    def __init__(
        self,
        model: nn.Module,
        target_module: str,
        device: int = None,
    ):
        """An Operation that runs a forward pass over each example in the
        dataset and stores model activations in a new column.

        Args:
            model (nn.Module): the torch model from which activations are extracted
            target_module (str): the name of the submodule of `model` (i.e. an
                intermediate layer) that outputs the activations we'd like to extract.
                For nested submodules, specify a path separated by "." (e.g.
                `ActivationCachedOp(model, "block4.conv")`).
            device (int, optional): the device for the forward pass. Defaults to None,
                in which case the CPU is used.
        """
        self.model = model
        self.device = device

        try:
            target_module = _nested_getattr(model, target_module)
        except nn.modules.module.ModuleAttributeError:
            raise ValueError(f"`model` does not have a submodule {target_module}")

        self.extractor = ActivationExtractor()
        target_module.register_forward_hook(self.extractor.add_hook)

        if self.device is not None:
            self.model.to(self.device)

        super(ActivationOp, self).__init__()

    def process_batch(
        self,
        dp: DataPanel,
        columns: List[str],
        **kwargs,
    ) -> tuple:
        """Compute example activations for a torch Model.

        Args:
            dp (DataPanel): DataPanel
            columns (list): list of columns
            **kwargs: optional keyword arguments

        Returns:
            Tuple with single output, a torch.Tensor.
        """

        inputs = dp[columns[0]]
        if self.device is not None:
            inputs = inputs.to(self.device)
        self.model(inputs)
        return (self.extractor.activation.cpu().detach(),)



from typing import Union

import torch


class DeviceMixin:
    def __init__(
        self,
        device: Union[int, str],
        *args,
        **kwargs,
    ):
        super(DeviceMixin, self).__init__(*args, **kwargs)

        if isinstance(device, str):
            if device.startswith("cuda"):
                self._cuda_device = (
                    0 if ":" not in device else int(device.split(":")[-1])
                )
            elif device.startswith("gpu"):
                self._cuda_device = (
                    0 if ":" not in device else int(device.split(":")[-1])
                )
            elif device.startswith("cpu"):
                self._cuda_device = None
        elif isinstance(device, int):
            assert device >= 0, "`device` must be >= 0."
            self._cuda_device = device
        else:
            self._cuda_device = -1
            self._device = "cuda" if torch.cuda.is_available() else "cpu"

        if self._cuda_device == -1:
            self._device = "cuda"
        elif self._cuda_device >= 0:
            self._device = f"cuda:{self._cuda_device}"
        else:
            self._device = "cpu"

    @property
    def cuda_device(self):
        return self._device

    @property
    def device(self):
        return self._device



from __future__ import annotations

from collections import Counter
from typing import List, Tuple

import numpy as np

from robustnessgym.core.identifier import Identifier
from robustnessgym.core.slice import SliceDataPanel as DataPanel
from robustnessgym.slicebuilders.subpopulation import Subpopulation


class EntityFrequency(Subpopulation):
    def __init__(self, entity_thresholds: List[Tuple[str, List[int]]], *args, **kwargs):

        identifiers = []
        for entity_type, thresholds in entity_thresholds:
            for threshold in thresholds:
                identifiers.append(
                    Identifier(
                        _name=self.__class__.__name__,
                        entity_type=entity_type,
                        threshold=threshold,
                    )
                )

        super(EntityFrequency, self).__init__(identifiers, *args, **kwargs)

        if len(entity_thresholds) == 0:
            raise ValueError("At least one entity type required")

        for entity_type, _ in entity_thresholds:
            if entity_type not in [
                "PERSON",
                "NORP",
                "FAC",
                "ORG",
                "GPE",
                "LOC",
                "PRODUCT",
                "EVENT",
                "WORK_OF_ART",
                "LAW",
                "LANGUAGE",
                "DATE",
                "TIME",
                "PERCENT",
                "MONEY",
                "QUANTITY",
                "ORDINAL",
                "CARDINAL",
            ]:
                raise ValueError(f"Invalid entity type: {entity_type}")

        # List of tuples, each of which contains an entity type and a list of
        # associated frequency thresholds
        self.entity_thresholds = entity_thresholds

    def apply(
        self,
        batch: DataPanel,
        columns: List[str],
        slice_membership: np.ndarray = None,
        *args,
        **kwargs,
    ) -> np.ndarray:

        if len(columns) != 1:
            raise ValueError("Only one key allowed")
        key = columns[0]

        for i, cache_item in enumerate(batch["cache"]):
            entities = cache_item["Spacy"][key]["ents"]
            entity_types = [ent["label"] for ent in entities]
            counts = Counter(entity_types)
            slice_ndx = 0
            for entity_type, thresholds in self.entity_thresholds:
                for threshold in thresholds:
                    if counts[entity_type] >= threshold:
                        slice_membership[i, slice_ndx] = 1
                    slice_ndx += 1

        return slice_membership



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# Note: To use the 'upload' functionality of this file, you must:
#   $ pipenv install twine --dev

import io
import os
import sys
from distutils.util import convert_path
from shutil import rmtree

from setuptools import Command, find_packages, setup

main_ns = {}
ver_path = convert_path("robustnessgym/version.py")
with open(ver_path) as ver_file:
    exec(ver_file.read(), main_ns)


# Package meta-data.
NAME = "robustnessgym"
DESCRIPTION = "Robustness Gym is an evaluation toolkit for machine learning."
URL = "https://github.com/robustness-gym/robustness-gym"
EMAIL = "kgoel@cs.stanford.edu"
AUTHOR = "The Robustness Gym Team"
REQUIRES_PYTHON = ">=3.7.0"
VERSION = main_ns["__version__"]

# What packages are required for this module to be executed?
REQUIRED = [
    "dataclasses>=0.6",
    "pandas",
    "meerkat-ml==0.1.2",
    "dill>=0.3.3",
    "numpy>=1.18.0",
    "cytoolz",
    "ujson",
    "transformers",
    "fastBPE>=0.1.0",
    "jsonlines>=1.2.0",
    "pyahocorasick>-1.4.0",
    "python-Levenshtein>=0.12.0",
    "torch>=1.8.0",
    "tqdm>=4.49.0",
    "datasets>=1.4.1",
    "PyYAML>=5.4.1",
    "omegaconf>=2.0.5",
    "fuzzywuzzy>=0.18.0",
    "semver>=2.13.0",
    "multiprocess>=0.70.11",
    "Cython>=0.29.21",
    "progressbar>=2.5",
    "plotly>=4.14.1",
    "kaleido>=0.1.0",
    "pytorch-lightning>=1.1.2",
    "sklearn",
]

# What packages are optional?
EXTRAS = {
    "dev": [
        "black>=21.5b0",
        "isort>=5.7.0",
        "flake8>=3.8.4",
        "docformatter>=1.4",
        "pytest-cov>=2.10.1",
        "sphinx-rtd-theme>=0.5.1",
        "nbsphinx>=0.8.0",
        "recommonmark>=0.7.1",
        "pre-commit>=2.9.3",
        "sphinx-autobuild",
        "twine",
    ],
    "augmentation": [
        "nlpaug>=1.1.1",
    ],
    "adversarial": [
        "textattack",
    ],
    "text": [
        "nltk",
        "textblob",
        "spacy",
        "stanza",
        "allennlp",
        "allennlp-models",
    ],
    "jupyter": [
        "meerkat-ml==0.1.2[jupyter]",
    ],
    "interactive": ["meerkat-ml==0.1.2[interactive]"],
    "summarization": [
        "rouge-score>=0.0.4",
    ],
    "ludwig": [
        "ludwig",
        "tensorflow",
    ],
    "tabular": [
        "meerkat-ml==0.1.2[tabular]",
    ],
    "vision": [
        "meerkat-ml==0.1.2[vision]",
    ],
}
EXTRAS["all"] = [e for l in EXTRAS.values() for e in l]

# The rest you shouldn't have to touch too much :)
# ------------------------------------------------
# Except, perhaps the License and Trove Classifiers!
# If you do change the License, remember to change the Trove Classifier for that!

here = os.path.abspath(os.path.dirname(__file__))

# Import the README and use it as the long-description.
# Note: this will only work if 'README.md' is present in your MANIFEST.in file!
try:
    with io.open(os.path.join(here, "README.md"), encoding="utf-8") as f:
        long_description = "\n" + f.read()
except FileNotFoundError:
    long_description = DESCRIPTION

# Load the package's __version__.py module as a dictionary.
about = {}
if not VERSION:
    project_slug = NAME.lower().replace("-", "_").replace(" ", "_")
    with open(os.path.join(here, project_slug, "__version__.py")) as f:
        exec(f.read(), about)
else:
    about["__version__"] = VERSION


class UploadCommand(Command):
    """Support setup.py upload."""

    description = "Build and publish the package."
    user_options = []

    @staticmethod
    def status(s):
        """Prints things in bold."""
        print("\033[1m{0}\033[0m".format(s))

    def initialize_options(self):
        pass

    def finalize_options(self):
        pass

    def run(self):
        try:
            self.status("Removing previous builds…")
            rmtree(os.path.join(here, "dist"))
        except OSError:
            pass

        self.status("Building Source and Wheel (universal) distribution…")
        os.system("{0} setup.py sdist bdist_wheel --universal".format(sys.executable))

        self.status("Uploading the package to PyPI via Twine…")
        os.system("twine upload dist/*")

        self.status("Pushing git tags…")
        os.system("git tag v{0}".format(about["__version__"]))
        os.system("git push --tags")

        sys.exit()


# Where the magic happens:
setup(
    name=NAME,
    version=about["__version__"],
    description=DESCRIPTION,
    long_description=long_description,
    long_description_content_type="text/markdown",
    author=AUTHOR,
    author_email=EMAIL,
    python_requires=REQUIRES_PYTHON,
    url=URL,
    packages=find_packages(exclude=["tests", "*.tests", "*.tests.*", "tests.*"]),
    # If your package is a single module, use this instead of 'packages':
    # py_modules=['mypackage'],
    # entry_points={
    #     'console_scripts': ['mycli=mymodule:cli'],
    # },
    install_requires=REQUIRED,
    extras_require=EXTRAS,
    include_package_data=True,
    license="Apache 2.0",
    classifiers=[
        # Trove classifiers
        # Full list: https://pypi.python.org/pypi?%3Aaction=list_classifiers
        "Programming Language :: Python",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.6",
        "Topic :: Scientific/Engineering :: Artificial Intelligence",
    ],
    # $ setup.py publish support.
    cmdclass={
        "upload": UploadCommand,
    },
)



# flake8: noqa
from .attack import Attack
from .slicebuilder import SliceBuilder
from .subpopulation import Subpopulation
from .subpopulation_collection import SubpopulationCollection
from .transformation import Transformation



from typing import List, Optional, Tuple

import cytoolz as tz
import numpy as np
from meerkat.provenance import capture_provenance

from robustnessgym.core.constants import SUBPOPULATION
from robustnessgym.core.identifier import Identifier
from robustnessgym.core.slice import SliceDataPanel as DataPanel
from robustnessgym.core.tools import strings_as_json
from robustnessgym.slicebuilders.slicebuilder import SliceBuilder


class Subpopulation(SliceBuilder):
    def __init__(self, identifiers: List[Identifier], apply_fn=None, *args, **kwargs):
        super(Subpopulation, self).__init__(
            identifiers=identifiers,
            category=SUBPOPULATION,
            apply_fn=apply_fn,
            *args,
            **kwargs,
        )

    def apply(
        self,
        batch: DataPanel,
        columns: List[str],
        slice_membership: np.ndarray = None,
        *args,
        **kwargs,
    ) -> np.ndarray:
        return NotImplemented("Must return a np.ndarray matrix.")

    def process_batch(
        self,
        dp: DataPanel,
        columns: List[str],
        *args,
        **kwargs,
    ) -> Tuple[Optional[List[DataPanel]], Optional[np.ndarray]]:

        # Construct the matrix of slice labels: (batch_size x num_slices)
        slice_membership = np.zeros((len(dp), self.num_slices), dtype=np.int32)

        # Apply the SliceBuilder's core functionality
        try:
            slice_membership = self.apply(
                dp, columns, slice_membership, *args, **kwargs
            )
        except TypeError:
            slice_membership = np.array(
                self.apply(dp, columns, *args, **kwargs)
            ).reshape((-1, self.num_slices))

        assert slice_membership.shape == (len(dp), self.num_slices), (
            "Output of apply has the wrong shape. "
            f"Expected: {(len(dp), self.num_slices)} and got: {slice_membership.shape}."
        )

        return None, slice_membership

    @capture_provenance(capture_args=["self", "columns", "batch_size"])
    def process_dataset(
        self,
        dp: DataPanel,
        columns: List[str],
        batch_size: int = 32,
        num_proc: int = None,
        *args,
        **kwargs,
    ) -> Tuple[List[DataPanel], np.ndarray]:

        # Create slices using the dataset
        all_slice_memberships = []

        # Batch the dataset, and process each batch
        for batch in dp.batch(batch_size):
            # Process the batch
            _, slice_memberships = self.process_batch(
                dp=batch,
                columns=columns,
                *args,
                **kwargs,
            )

            # Keep track of the slice memberships
            all_slice_memberships.append(slice_memberships)

        # Create a single slice label matrix
        slice_membership = np.concatenate(all_slice_memberships, axis=0)

        slices = []
        for i in range(len(self.identifiers)):
            # Create a view of the original DataPanel
            sl = dp.view()

            # Only keep the filtered rows visible
            for column in sl._data.values():
                column.visible_rows = np.where(slice_membership[:, i])[0]

            # Set the Slice category using the SliceBuilder's category
            sl.category = self.category

            # Append the the lineage
            sl.add_to_lineage(
                category=str(self.category.capitalize()),
                identifier=self.identifiers[i],
                columns=strings_as_json(columns),
            )

            #
            # sl.identifier = ...

            slices.append(sl)

        # for i, sl in enumerate(slices):
        #     # Set the visible rows for each slice
        #     sl.set_visible_rows(np.where(slice_membership[:, i])[0])

        return slices, slice_membership

    @classmethod
    def union(
        cls, *slicebuilders: SliceBuilder, identifier: Identifier = None
    ) -> SliceBuilder:
        """Combine a list of slicers using a union."""
        # Group the slicers based on their class
        grouped_slicers = tz.groupby(lambda s: s.__class__, slicebuilders)

        # Join the slicers corresponding to each class, and flatten
        slicebuilders = list(
            tz.concat(
                tz.itemmap(
                    lambda item: (item[0], item[0].join(*item[1])), grouped_slicers
                ).values()
            )
        )

        def apply_fn(slice_membership, batch, columns, *args, **kwargs):
            # Determine the size of the batch
            batch_size = len(batch[list(batch.keys())[0]])

            # Keep track of all the slice labels
            all_slice_membership = []

            # Run each slicebuilder on the batch
            for slicebuilder in slicebuilders:
                all_slice_membership.append(
                    slicebuilder.apply(
                        slice_membership=np.zeros(
                            (batch_size, slicebuilder.num_slices), dtype=np.int32
                        ),
                        batch=batch,
                        columns=columns,
                    )
                )

            # Concatenate all the slice labels
            slice_membership = np.concatenate(all_slice_membership, axis=1)

            # Take the union over the slices (columns)
            slice_membership = np.any(slice_membership, axis=1).astype(np.int32)[
                :, np.newaxis
            ]

            return slice_membership

        return Subpopulation(identifiers=[identifier], apply_fn=apply_fn)

    @classmethod
    def intersection(
        cls, *slicebuilders: SliceBuilder, identifier: Identifier = None
    ) -> SliceBuilder:
        """Combine a list of slicemakers using an intersection."""
        # Group the slicemakers based on their class
        grouped_slicemakers = tz.groupby(lambda s: s.__class__, slicebuilders)

        # Join the slicemakers corresponding to each class, and flatten
        slicebuilders = list(
            tz.concat(
                tz.itemmap(
                    lambda item: (item[0], item[0].join(*item[1])), grouped_slicemakers
                ).values()
            )
        )

        def apply_fn(slice_membership, batch, columns, *args, **kwargs):
            # Determine the size of the batch
            batch_size = len(batch[list(batch.keys())[0]])

            # Keep track of all the slice labels
            all_slice_membership = []

            # Run each slicemaker on the batch
            for slicebuilder in slicebuilders:
                all_slice_membership.append(
                    slicebuilder.apply(
                        slice_membership=np.zeros(
                            (batch_size, slicebuilder.num_slices), dtype=np.int32
                        ),
                        batch=batch,
                        columns=columns,
                    )
                )

            # Concatenate all the slice labels
            slice_membership = np.concatenate(all_slice_membership, axis=1)

            # Take the union over the slices (columns)
            slice_membership = np.all(slice_membership, axis=1).astype(np.int32)[
                :, np.newaxis
            ]

            return slice_membership

        return Subpopulation(identifiers=[identifier], apply_fn=apply_fn)
